# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import math
import os
import random
import re
import signal
import time
from threading import Thread
from collections.abc import Sequence
from contextlib import contextmanager
from itertools import count
from numbers import Number

from .backends import make_backend
from .devices import Bulb, Beacon
from .robot import Robot
from .utils import (
    Food,
    Grid,
    Line,
    Point,
    cast_ray,
    distance,
    distance_point_to_line,
    format_time,
    in_colab,
    json_dump,
    load_image,
    rotate_around,
    print_once,
    PI_OVER_180,
    PI_OVER_2,
    ONE80_OVER_PI,
    TWO_PI,
)

from aitk.utils import Color, progress_bar

BLACK = Color("black")
BLUE = Color("blue")

try:
    from IPython.display import display, clear_output
except ImportError:
    display = print
    def clear_output(wait=True):
        pass

DEFAULT_HANDLER = signal.getsignal(signal.SIGINT)


class Wall:
    """
    Class representing obstacles in the world. If the bounding box of
    a robot, then the parameter robot will be that robot, else None.
    """

    def __init__(self, color, robot, *lines, wtype=None):
        """
        Create a wall consisting of one or more lines.

        wtype is "boundary", "robot", or "wall"
        """
        self.color = color
        self.robot = robot
        self.lines = lines
        self.wtype = wtype

    def __repr__(self):
        return "Wall(%r, %r, %r, wtype=%r)" % (self.color, self.robot, self.lines, self.wtype)


class List(Sequence):
    def __init__(self, list_object):
        self.list_object = list_object

    def __getattr__(self, attr):
        return getattr(self.list_object, attr)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.list_object[item]
        elif isinstance(item, str):
            name_map = {}  # mapping of base to count
            search_groups = re.match(r"(.*)-(\d*)", item)
            if search_groups:
                search_name = search_groups[1].lower()
                if search_groups[2].isdigit():
                    search_index = int(search_groups[2])
                else:
                    search_name = item
                    search_index = 1
            else:
                search_name = item.lower()
                search_index = 1
            for item in self.list_object:
                # update name_map
                name = item.name.lower()
                index = None
                if "-" in name:
                    prefix, index = name.rsplit("-", 1)
                    if index.isdigit():
                        name = prefix
                        index = int(index)
                    else:
                        index = 1
                if name not in name_map:
                    name_map[name] = 1
                else:
                    name_map[name] += 1
                if index is None:
                    index = name_map[name]
                if search_name == name and search_index == index:
                    return item

        return None

    def __len__(self):
        return len(self.list_object)

    def __repr__(self):
        return repr(self.list_object)

class CanvasCall:
    def __init__(self, canvas, command):
        self.canvas = canvas
        self.command = command

    def __call__(self, *args, **kwargs):
        self.canvas.command_list.append((self.command, args, kwargs))

class Canvas:
    def __init__(self, command_list):
        self.command_list = command_list

    def __getattr__(self, attr):
        return CanvasCall(self, attr)

    def clear(self):
        self.command_list.clear()

class World:
    """
    The aitk.robots simulator world. All simulations
    begin with a world.
    """

    def __init__(
        self,
        width=500, # type: int
        height=250,  # type: int
        seed=0,  # type: int
        scale=3.0, # type: float
        boundary_wall=True,  # type: bool
        boundary_wall_color="purple",  # type: str
        boundary_wall_width=1,  # type: int
        ground_color="green",  # type: str
        ground_image_filename=None,  # type: str
        filename=None,  # type: str
        quiet=False,  # type: bool
        smell_cell_size=None,  # type: int
        **kwargs
    ):
        """
        The constructor for the aitk.robots simulator world.

        Args:
            width (int): width of world in pixels
            height (int): height of world in pixels
            seed (int): random number generator seed
            scale (float): value to use in drawing the world
            boundary_wall (bool): draw a boundary around the world?
            boundary_wall_color (str): color name of boundary wall
            boundary_wall_width (int): width of boundary wall
            ground_color (str): color name
            ground_image_filename (str): image file used for backgound
            filename (str): name of json world file
            quiet (bool): if True, don't print any messages
            smell_cell_size (int): is set automatically based on world size

        You can also pass any valid item from the world config settings.

        ```python
        from aitk.robots import World

        world = World(200, 200)
        ```
        """
        # For faster-than real time display with synchronous backends,
        # keep processing time below this percentage of throttle_period:
        config = {
            "width": width,
            "height": height,
            "seed": seed,
            "scale": scale,
            "boundary_wall": boundary_wall,
            "boundary_wall_width": boundary_wall_width,
            "boundary_wall_color": boundary_wall_color,
            "ground_color": ground_color,
            "quiet": quiet,
            "smell_cell_size": smell_cell_size,
        }
        if filename is not None:
            config["filename"] = filename
        if ground_image_filename is not None:
            config["ground_image_filename"] = ground_image_filename
        config["walls"] = kwargs.pop("walls", [])
        config["bulbs"] = kwargs.pop("bulbs", [])
        config["beacon"] = kwargs.pop("beacon", None)
        config["robots"] = kwargs.pop("robots", [])
        config["food"] = kwargs.pop("food", [])
        if len(kwargs) != 0:
            raise AttributeError(
                "unknown arguments for World: %s" % list(kwargs.keys())
            )
        self.time_step = 0.10  # seconds
        self._events = []
        self._show_throttle_percentage = 0.40
        self._time_decimal_places = 1
        self._throttle_period = 0.1
        self._time_of_last_call = 0
        self._step_display = "tqdm"
        self.debug = False
        self._watchers = []
        self._robots = []
        self._bulbs = []
        self._beacon = None
        self._backend = None
        self._recording = False
        self.config = config.copy()
        self._initialize()  # default values
        self.robots = List(self._robots)
        self.bulbs = List(self._bulbs)
        self.reset()  # from config

    def __repr__(self):
        return "<World width=%r, height=%r>" % (self.width, self.height)

    def _repr_png_(self):
        return self._backend.to_image("png")

    def _event(self, etype, **kwargs):
        if etype == "food-off":
            robot = kwargs["robot"]
            food = kwargs["food"]
            robot.food_eaten += 1
            food.state = "off"
            self._grid.need_update = True
            robot_index = self._robots.index(robot)
            food_index = self._food.index(food)
            self._events.append((self.time, "food-off", robot_index, food_index))
        elif etype == "bulb-on":
            bulb = kwargs["bulb"]
            bulb_index = self._bulbs.index(bulb)
            self._events.append((self.time, "bulb-on", bulb_index))
        elif etype == "bulb-color":
            bulb = kwargs["bulb"]
            previous_color = kwargs["previous_color"]
            new_color = kwargs["new_color"]
            bulb_index = self._bulbs.index(bulb)
            self._events.append((self.time, "bulb-color", bulb_index, previous_color, new_color))
        elif etype == "bulb-off":
            bulb = kwargs["bulb"]
            bulb_index = self._bulbs.index(bulb)
            self._events.append((self.time, "bulb-off", bulb_index))
        elif etype == "beacon-on":
            self._events.append((self.time, "beacon-on"))
        elif etype == "beacon-off":
            self._events.append((self.time, "beacon-off"))
        else:
            raise Exception("unknown event: %s" % etype)
        self.update() # request draw


    def get_time(self):
        """
        Get the simulated time as a formatted string.
        """
        return format_time(self.time)

    def get_image(self, index=None, size=100, copy=True):
        """
        Get a PIL image (copy) of the world, or of a robot.

        Args:
            index (str or int, optional): index of robot
            size (int, optional): size of robot picture
            copy (bool, optional): get a copy of image if True,
                otherwise return a reference to world image
        """
        try:
            picture = self._backend.get_image(self.time)
        except RuntimeError:
            raise Exception("Backend is not ready yet; try again")

        if index is not None:
            robot = self.robots[index]
            if robot:
                start_x = round(max(robot.x * self.scale - size / 2, 0))
                start_y = round(max(robot.y * self.scale - size / 2, 0))
                rectangle = (
                    start_x,
                    start_y,
                    min(start_x + size, self.width * self.scale),
                    min(start_y + size, self.height * self.scale),
                )
                picture = picture.crop(rectangle)
                return picture
        else:
            if copy:
                picture = picture.copy()
            return picture

    def display(self, index=None, size=100):
        """
        Display a picture of the world, or of a robot.

        Args:
            index (str or int, optional): index of robot
            size (int, optional): size of robot picture
        """
        picture = self.get_image(index=index, size=size)
        display(picture)

    def summary(self):
        """
        Print a summary of information about the world and
        all of its robots.
        """
        print("World details:")
        if self.filename:
            print("This world was loaded from %r" % self.filename)
        print("Size: %s x %s" % (self.width, self.height))
        print("-" * 25)
        print("Robots:")
        print("-" * 25)
        if len(self._robots) == 0:
            print("  This world has no robots.")
        else:
            for i, robot in enumerate(self._robots):
                print("  .robots[%s or %r]: %r" % (i, robot.name, robot))
                robot.summary()
        print("-" * 25)
        print("Food:")
        print("-" * 25)
        if len(self._food) == 0:
            print("  This world has no food.")
        else:
            for food in self._food:
                print("  x: %s, y: %s, brightness: %s, state: %s" % (
                    food.x, food.y, food.standard_deviation, food.state))
        print("-" * 25)
        print("Lights:")
        print("-" * 25)
        if len(self._bulbs) == 0:
            print("  This world has no lights.")
        else:
            for bulb in self._bulbs:
                print("  x: %s, y: %s, brightness: %s, name: %r, color: %s" % (
                    bulb.x,
                    bulb.y,
                    bulb.brightness,
                    bulb.name,
                    bulb.color))
        print("-" * 25)
        print("Beacon:")
        print("-" * 25)
        if self.beacon is None:
            print("  This world has no beacon.")
        else:
            x, y = self.beacon.get_position(world=True)
            print("  x: %s, y: %s" % (x, y))

    def get_robot(self, item):
        """
        Get the robot by name or index. Equivalent to
        world.robots[item]

        Args:
            item (int or string): index or name of robot
        """
        return self.robots[item]

    def get_bulb(self, item):
        """
        Get the bulb by name or index. Equivalent to
        world.bulbs[item]

        Args:
            item (int or string): index or name of bulb
        """
        return self.bulbs[item]

    @property
    def beacon(self):
        """
        Get the beacon.
        """
        if self._beacon is not None:
            return self._beacon

        for robot in self.robots:
            for device in robot._devices:
                if device.type == "beacon":
                    return device
        return None

    @beacon.setter
    def beacon(self, value):
        self._beacon = value

    def _initialize(self):
        """
        Sets the default values.
        """
        self._draw_list = []
        self._overlay_list = []
        self.canvas = Canvas(self._overlay_list)
        self.filename = None
        self.quiet = False
        self.seed = 0
        self.width = 500
        self.height = 250
        self.scale = 3.0
        self._stop = False  # should stop?
        self._thread = None
        self.status = "stopped"
        self.time = 0.0  # seconds
        self.boundary_wall = True
        self.boundary_wall_width = 1
        self.boundary_wall_color = Color(128, 0, 128)
        self.ground_color = Color(0, 128, 0)
        self.ground_image_filename = None
        self._ground_image = None
        self._ground_image_pixels = None
        self._walls = []
        self._bulbs.clear()
        self._beacon = None
        self._complexity = 0
        self.smell_cell_size = None

    def reset(self):
        """
        Reloads the config from initialization, or from
        last save.
        """
        self.stop()
        self._initialize()
        self._reset_watchers()
        self._food = []
        self._events = []
        self._grid = None
        self.from_json(self.config)
        self.time = 0.0
        for robot in self._robots:
            robot.reset()
            # Re-add the robot's boundaries:
            wall = Wall(robot.color, robot, *robot._bounding_lines, wtype="robot")
            self._walls.append(wall)
        self._stop = False  # should stop?
        self.status = "stopped"
        self.update(show=False)  # twice to allow robots to see each other
        self.update(show=False)
        self.draw()  # force

    def set_seed(self, seed):
        """
        Set the random seed.
        """
        if seed == 0:
            seed = random.randint(0, 9999999)
            if not self.quiet:
                print("Random seed set to:", seed)
        else:
            if not self.quiet:
                print("Using random seed:", seed)
        random.seed(seed)
        self.seed = seed
        self.config["seed"] = seed

    def from_json(self, config):
        """
        Load a json config file.
        """
        self.config = config

        if "quiet" in config:
            self.quiet = config["quiet"]

        seed = config.get("seed", 0)
        self.set_seed(seed)

        if "filename" in config:
            self.filename = config["filename"]
        if "width" in config:
            self.width = config["width"]
        if "height" in config:
            self.height = config["height"]
        if "scale" in config:
            self.scale = config["scale"]
        if "smell_cell_size" in config:
            self.smell_cell_size = config["smell_cell_size"]

        if self.smell_cell_size is None:
            self.smell_cell_size = max((self.width * self.height) // 20000, 1)

        if "boundary_wall" in config:
            self.boundary_wall = config["boundary_wall"]
        if "boundary_wall_color" in config:
            self.boundary_wall_color = Color(config["boundary_wall_color"])
        if "boundary_wall_width" in config:
            self.boundary_wall_width = config["boundary_wall_width"]
        if "ground_color" in config:
            self.ground_color = Color(config["ground_color"])
        if "ground_image_filename" in config:
            self.set_ground_image(config["ground_image_filename"], show=False)

        # Now, we create the grid:
        self._grid = Grid(self.width, self.height, self.smell_cell_size)
        self._grid.update_walls(self._walls)

        # Add walls:
        self._add_boundary_walls()
        for wall in config.get("walls", []):
            # Walls are either "boxes" with 4 lines, or a single line:
            if "wtype" not in wall or wall["wtype"] == "box":
                self.add_wall(
                    wall["color"],
                    wall["p1"]["x"],
                    wall["p1"]["y"],
                    wall["p2"]["x"],
                    wall["p2"]["y"],
                    box=True,
                )
            elif wall["wtype"] == "line":
                self.add_wall(
                    wall["color"],
                    wall["p1"]["x"],
                    wall["p1"]["y"],
                    wall["p2"]["x"],
                    wall["p2"]["y"],
                    box=False,
                )

        beacon = config.get("beacon", None)
        if beacon is not None:
            # beacon is {x, y, z}
            self._add_beacon(**beacon)

        for bulb in config.get("bulbs", []):
            # bulbs are {x, y, z, color, brightness}
            self._add_bulb(**bulb)

        for food in config.get("food", []):
            # food x, y, standard_deviation, state
            self._add_food(**food)

        ## Create robot, and add to world:
        for i, robotConfig in enumerate(self.config.get("robots", [])):
            # FIXME: raise if lengths don't match
            if i < len(self._robots):  # already a robot; let's reuse it:
                robot = self._robots[i]
                robot._initialize()
                robot.from_json(robotConfig)
            else:
                robot = Robot(**robotConfig)
                self.add_robot(robot)
        # Create the backend if first time:
        if self._backend is None:
            self._backend = make_backend(self.width, self.height, self.scale)
        # Update the backend if it already existed, but differs in config
        self._backend.update_dimensions(self.width, self.height, self.scale)

    def add_food(self, x, y, standard_deviation, state="on"):
        """
        Add food at x, y with a brightness of standard_deviation
        (in pixels).

        Args:
            x (int): the x coordinate
            y (int): the y coordinate
            standard_deviation (float): distance int pixels
                of 1 standard deviation
        """
        self._add_food(x, y, standard_deviation, state)
        self._grid.need_update = True
        self.update()  # request draw
        self.save()

    def _add_food(self, x, y, standard_deviation, state="on"):
        self._food.append(Food(x, y, standard_deviation, state))

    def get_food(self):
        """
        Return a list of the food in the world in
        (x, y, standard_deviation, state) form.
        """
        return self._food

    def _add_boundary_walls(self):
        """
        Add boundary walls around world.
        """
        if self.boundary_wall:
            p1 = Point(0, 0)
            p2 = Point(0, self.height)
            p3 = Point(self.width, self.height)
            p4 = Point(self.width, 0)
            ## Not a box, but surround area with four boundaries:
            self._walls.extend(
                [
                    Wall(self.boundary_wall_color, None, Line(p1, p2), wtype="boundary"),
                    Wall(self.boundary_wall_color, None, Line(p2, p3), wtype="boundary"),
                    Wall(self.boundary_wall_color, None, Line(p3, p4), wtype="boundary"),
                    Wall(self.boundary_wall_color, None, Line(p4, p1), wtype="boundary"),
                ]
            )
            self._complexity = self._compute_complexity()

    def to_json(self):
        """
        Get the world as a JSON dict.
        """
        config = {
            "seed": self.seed,
            "width": int(self.width),
            "height": int(self.height),
            "scale": self.scale,
            "boundary_wall": self.boundary_wall,  # bool
            "boundary_wall_color": str(self.boundary_wall_color),
            "boundary_wall_width": self.boundary_wall_width,
            "ground_color": str(self.ground_color),
            "ground_image_filename": self.ground_image_filename,
            "quiet": self.quiet,
            "smell_cell_size": self.smell_cell_size,
            "walls": [],
            "bulbs": [],
            "beacon": None,
            "robots": [],
            "food": [],
        }
        if self._beacon is not None:
            config["beacon"] = {
                "x": self._beacon.x,
                "y": self._beacon.y,
                "z": self._beacon.z,
            }
        for wall in self._walls:
            # Not a boundary wall or robot bounding box:
            if wall.wtype == "wall":
                if len(wall.lines) == 4:
                    # Box:
                    w = {
                        "color": str(wall.color),
                        "p1": {"x": wall.lines[0].p1.x, "y": wall.lines[0].p1.y,},
                        "p2": {"x": wall.lines[2].p1.x, "y": wall.lines[2].p1.y,},
                        "wtype": "box",
                    }
                elif len(wall.lines) == 1:
                    # Line:
                    w = {
                        "color": str(wall.color),
                        "p1": {"x": wall.lines[0].p1.x, "y": wall.lines[0].p1.y,},
                        "p2": {"x": wall.lines[0].p2.x, "y": wall.lines[0].p2.y,},
                        "wtype": "line",
                    }
                else:
                    raise Exception("invalid wall length; should be 1 or 4: %s" % len(wall.lines))
                config["walls"].append(w)

        for bulb in self._bulbs:
            config["bulbs"].append(
                {
                    "color": str(bulb.color),
                    "x": bulb.x,
                    "y": bulb.y,
                    "z": bulb.z,
                    "brightness": bulb.brightness,
                }
            )

        for food in self._food:
            config["food"].append(
                {
                    "x": food.x,
                    "y": food.y,
                    "standard_deviation": food.standard_deviation,
                    "state": food.state,
                }
            )

        for robot in self._robots:
            config["robots"].append(robot.to_json())

        return config

    def save(self):
        """
        Save the current state of the world as the config.
        """
        self.config = self.to_json()

    def save_file(self):
        """
        Save the current state of the world as the config, and
        save it back to disc if it was loaded from disk.
        """
        # First, save internally.
        self.config = self.to_json()
        if self.filename is not None and os.path.exists(self.filename):
            with open(self.filename, "w") as fp:
                json_dump(self.config, fp, sort_keys=True, indent=4)
        else:
            if not self.quiet:
                print("Saved in memory. Use world.save_as('filename') to save to disk.")

    def save_as(self, filename):
        """
        Save the world config JSON as a new file.

        Args:
            filename (str): the name of the file to save
                the world in
        """
        if not filename.endswith(".json"):
            filename = filename + ".json"
        # First, save internally.
        self.config = self.to_json()
        with open(filename, "w") as fp:
            json_dump(self.to_json(), fp, sort_keys=True, indent=4)
        self.config["filename"] = filename
        # Now you can use save():

    def set_ground_image(self, filename, show=True):
        """
        Set the background image

        Args:
            filename (str): the name of the image
            show (bool): update the world now if True
        """
        self.ground_image_filename = filename
        self._reset_ground_image()
        if show:
            self.update(show=False)
            self.draw()  # force

    def _reset_ground_image(self):
        """
        Reset the ground image, in case it changed.
        """
        if self.ground_image_filename is not None:
            self._ground_image = load_image(
                self.ground_image_filename,
                round(self.width * self.scale),
                round(self.height * self.scale),
            )
        else:
            self._ground_image = None

        if self._ground_image is not None:
            self._ground_image_pixels = self._ground_image.load()

    def paste_ground_image(self, image, x, y):
        """
        Paste an image onto the ground image. Requires
        a ground image to have already been set.

        Args:
            image (Image): a Python Image Library image
            x (int): the x coordinate of upper lefthand corner
            y (int): the y coordinate of upper lefthand corner
        """
        if self._ground_image:
            self._ground_image.paste(image, (x, y))

    def set_ground_color_at(self, x, y, pen):
        """
        Set the pixel(s) of the ground image at position (x,y). Requires
        a ground image to have already been set.

        Args:
            x (int): the x coordinate
            y (int): the y coordinate
            pen (tuple): the (color, radius) to draw with
        """
        if self._ground_image:
            color, radius = pen
            for i in range(-radius, radius + 1, 1):
                for j in range(-radius, radius + 1, 1):
                    try:
                        self._ground_image_pixels[
                            ((x * self.scale) + i, (y * self.scale) + j)
                        ] = color.to_tuple()
                    except Exception:
                        pass

    def get_ground_color_at(self, x, y, radius=1):
        """
        Get the pixel(s) of the ground image at position (x,y). Requires
        a ground image to have already been set.

        Args:
            x (int): the x coordinate
            y (int): the y coordinate
            radius (int): size of area
        """
        if self._ground_image:
            results = []
            for i in range(-radius, radius + 1, 1):
                for j in range(-radius, radius + 1, 1):
                    results.append(
                        self._ground_image_pixels[
                            ((x + i) * self.scale, (y + j) * self.scale)
                        ]
                    )
            return results

    def set_scale(self, scale):
        """
        Change the scale of the rendered world.

        Args:
            scale (float): the scale of the world (usually between 0.5 and 10)
        """
        self.scale = scale
        self._backend.update_dimensions(self.width, self.height, self.scale)
        # Save with config
        self.config["scale"] = self.scale
        self.update(show=False)
        self.draw()  # force

    def watch(self, width=None, height=None):
        """
        Display a live (dynamically updated) view of the world.

        Args:
            width (int): the width of the world
            height (int): the height of the world
        """
        # Force update:
        self.update(show=True)
        # Force draw:
        self.draw()
        widget = self.get_widget(width, height)
        display(widget)

    def get_widget(self, width=None, height=None):
        """
        Returns the IPython simulation widget.

        Args:
            width (int): the width of the world
            height (int): the height of the world
        """
        if isinstance(width, int):
            width = "%spx" % width
        if isinstance(height, int):
            height = "%spx" % height
        self._step_display = "notebook"
        self.update()
        return self._backend.get_widget(width=width, height=height)

    def record(self):
        """
        Create and return a world recorder.
        """
        from .watchers import Recorder

        recorder = Recorder(self)
        self._watchers.append(recorder)
        self._recording = True
        return recorder

    def _draw_watchers(self):
        if self._backend is not None:
            self._backend.draw_watcher()
        for watcher in self._watchers:
            watcher.draw()

    def _reset_watchers(self):
        if self._backend is not None:
            self._backend.reset_watcher()
        for watcher in self._watchers:
            watcher.reset()

    def _update_watchers(self):
        if self._backend is not None:
            self._backend.update_watcher()
        for watcher in self._watchers:
            watcher.update()

    def clear_watchers(self):
        """
        Clear all of the watchers.
        """
        self._watchers[:] = []

    def add_bulb(self, color, x, y, z, brightness, name=None):
        """
        Add a bulb to the world.

        Args:
            color (str): the color of the light
            x (int): the x coordinate
            y (int): the y coordinate
            z (int): the z coordinate
            brightness (float): the distance in pixels of 1 standard deviation
            name (str): the name of the bulb
        """
        self._add_bulb(color, x, y, z, brightness, name)
        self.update()  # request draw
        self.save()

    def _add_bulb(self, color, x, y, z, brightness, name=None):
        name = name if name is not None else "bulb-%s" % (len(self._bulbs) + 1)
        bulb = Bulb(color, x, y, z, brightness, name, self)
        self._bulbs.append(bulb)

    def add_beacon(self, x, y, z=0):
        """
        Add a beacon to the world.

        Args:
            x (int): the x coordinate
            y (int): the y coordinate
            z (int): optional, the z coordinate
        """
        self._add_beacon(x, y, z)
        self.update()  # request draw
        self.save()

    def _add_beacon(self, x, y, z=0):
        beacon = Beacon(x, y, z, self)
        self._beacon = beacon

    def add_wall(self, color, x1, y1, x2, y2, box=True, wtype="wall"):
        """
        Add a wall as a line or as a box.

        Args:
            color (str): the color of the wall
            x1 (int): the upper left x coordinate
            y1 (int): the upper left y coordinate
            x2 (int): the lower right x coordinate
            y2 (int): the lower right y coordinate
            box (bool): whether this is a box (True) or line (False)
            wtype (str): "wall", "robot", or "boundary"
        """
        p1 = Point(x1, y1)
        p3 = Point(x2, y2)
        if box:
            ## Pairs of points make Line:
            p2 = Point(x2, y1)
            p4 = Point(x1, y2)
            wall = Wall(
                Color(color),
                None,
                Line(p1, p2),
                Line(p2, p3),
                Line(p3, p4),
                Line(p4, p1),
                wtype=wtype,
            )
        else:
            wall = Wall(
                Color(color), None, Line(p1, p3), wtype="wall"
            )
        self._walls.append(wall)
        self._complexity = self._compute_complexity()
        self._grid.update_wall(wall)
        self.draw()

    def del_robot(self, robot):
        """
        Remove a robot from the world.

        Args:
            robot (Robot): the robot to remove
        """
        if not isinstance(robot, Robot):
            # Then look it up by index/name/type:
            robot = self.robots[robot]
        for wall in list(self._walls):
            if wall.robot is robot:
                self._walls.remove(wall)
        if robot in self._robots:
            robot.world = None
            self._robots.remove(robot)
        self._complexity = self._compute_complexity()
        self.update()  # request draw

    def _is_enclosed(self, robot, x, y):
        # Determine if robot is contained in a wall box
        # cast two rays in opposite directions
        maxRange = max(self.width, self.height)
        hits1 = cast_ray(self, robot, x, y, 0, maxRange,
                         ignore_robots=[robot])
        hits2 = cast_ray(self, robot, x, y, math.pi, maxRange,
                         ignore_robots=[robot])
        # if both hit part of same wall, then contained
        return (len(hits1) > 0 and
                len(hits2) > 0 and
                hits1[-1].wall == hits2[-1].wall)

    def _find_random_pose(self, robot):
        """
        Add a robot to the world in a random position.
        """
        pa = random.random() * math.pi * 2
        for i in range(100):
            too_close = False
            # Propose a position:
            px = round(robot.radius + random.random() * (self.width - 2 * robot.radius))
            py = round(
                robot.radius + random.random() * (self.height - 2 * robot.radius)
            )
            # Check distance with other robots:
            for other in self._robots:
                if distance(px, py, other.x, other.y) < robot.radius + other.radius:
                    too_close = True
                    break

            # Check other non-robot walls:
            if not too_close:
                for wall in self._walls:
                    if wall.robot is not None:
                        continue
                    for line in wall.lines:
                        dist, location = distance_point_to_line((px, py), line.p1, line.p2)
                        if dist < robot.radius:
                            too_close = True
                            break
                    if too_close:
                        break
            # If it isn't too close, check if enclosed:
            if not too_close:
                if not self._is_enclosed(robot, px, py):
                    # If all good, return position
                    return px, py, pa

        raise Exception("Couldn't find a place for robot after 100 tries; giving up")

    def add_robot(self, robot):
        """
        Add a new robot to the world. If the robot's position is 0,0 then
        place it randomly in the world.

        Args:
            robot (Robot): the robot to add
        """
        if robot not in self._robots:
            if robot.x == 0 and robot.y == 0:
                robot.x, robot.y, robot.a = self._find_random_pose(robot)
            self._robots.append(robot)
            robot.world = self
            # Bounding lines form a wall:
            if len(robot._bounding_lines) == 0:
                print("WARNING: adding a robot with no body")
            wall = Wall(robot.color, robot, *robot._bounding_lines, wtype="robot")
            self._walls.append(wall)
            self._complexity = self._compute_complexity()
            self.update()
            self.save()
        else:
            raise Exception("Can't add the same robot to a world more than once.")

    def _signal_handler(self, *args, **kwargs):
        """
        Handler for Control+C.
        """
        self._stop = True

    @contextmanager
    def _no_interrupt(self):
        """
        Suspends signal handling execution
        """
        self._stop = False
        try:
            signal.signal(signal.SIGINT, self._signal_handler)
        except ValueError:
            # Cannot do this in a thread
            pass

        try:
            yield None
        finally:
            try:
                signal.signal(signal.SIGINT, DEFAULT_HANDLER)
            except ValueError:
                # Cannot do this in a thread
                pass

    def stop(self):
        """
        Stop the simulator, if in thread or not.
        """
        self._stop = True
        self.status = "stopped"
        if self._thread is not None:
            print("Stopping thread...")
            self._thread.join()
            self._thread = None

    def run(
        self,
        function=None,
        time_step=None,
        show=True,
        real_time=True,
        show_progress=True,
        quiet=False,
        background=False,
        interrupt=False,
        callback=None,
    ):
        """
        Run the simulator until one of the control functions returns True
        or Control+C is pressed.

        Args:
            function (Function): either a single function that takes the
                world, or a list of functions (or None) that each take
                a robot. If any function returns True, then simulation will
                stop.
            time_step (float): time unit to advance the world
            show (bool): update the watchers
            real_time (bool): run simulation in real time
            show_progress (bool): show progress bar
            quiet (bool): if True, do not show the status message when
                completed
            background (bool): if True, run in the background.
            interrupt (bool): if True, raise KeyboardInterrupt, otherwise
                just stop and continue
            callback: a function that takes the world to run on each step
        """
        time_step = time_step if time_step is not None else self.time_step
        if background:
            if self._thread is None:
                kwargs = {
                    "function": function,
                    "time_step": time_step,
                    "show": show,
                    "real_time": real_time,
                    "show_progress": False,
                    "quiet": True,
                    "background": False
                }
                print("Starting world.run() in background. Use world.stop()")
                self._thread = Thread(target=self.run, kwargs=kwargs)
                self._thread.start()
            else:
                print("The world is already running in the background. Use world.stop()")
        else:
            return self.steps(
                float("inf"), function, time_step, show, real_time,
                show_progress, quiet, interrupt, callback
            )

    def seconds(
        self,
        seconds=5.0,
        function=None,
        time_step=None,
        show=True,
        real_time=True,
        show_progress=True,
        quiet=False,
        interrupt=False,
        callback=None,
    ):
        """
        Run the simulator for N seconds, or until one of the control
        functions returns True or Control+C is pressed.

        Args:
            seconds (float): how many simulation seconds to run
            function (Function): either a single function that takes the
                world, or a list of functions (or None) that each take
                a robot. If any function returns True, then simulation will
                stop.
            time_step (float): time unit to advance the world
            show (bool): update the watchers
            real_time (float): run simulation in real time
            show_progress (bool): show progress bar
            quiet (bool): if True, do not show the status message when
                completed
            interrupt (bool): if True, raise KeyboardInterrupt, otherwise
                just stop and continue
            callback: a function that takes the world to run on each step
        """
        time_step = time_step if time_step is not None else self.time_step
        steps = round(seconds / time_step)
        return self.steps(steps, function, time_step, show, real_time,
                          show_progress, quiet, interrupt, callback)

    def steps(
        self,
        steps=1,
        function=None,
        time_step=None,
        show=True,
        real_time=True,
        show_progress=True,
        quiet=False,
        interrupt=False,
        callback=None,
    ):
        """
        Run the simulator for N steps, or until one of the control
        functions returns True or Control+C is pressed.

        Args:
            steps (int): either a finite number, or infinity
            function (Function) either a single function that takes the
                world, or a list of functions (or None) that each take
                a robot. If any function returns True, then simulation will
                stop.
            time_step (float): time unit to advance the world
            show (bool): update the watchers
            real_time (bool): run simulation in real time
            show_progress (bool): show progress bar
            quiet (bool): if True, do not show the status message when
                completed
            interrupt (bool): if True, raise KeyboardInterrupt, otherwise
                just stop and continue
            callback: a function that takes the world to run on each step
        """
        self.status = "running"
        stop_values = []
        time_step = time_step if time_step is not None else self.time_step
        if steps == float("inf"):
            step_iter = count()
        else:
            step_iter = range(steps)
        with self._no_interrupt():
            start_real_time = time.monotonic()
            start_time = self.time
            # If you want to show, and in google colab, then
            # we use tqdm's update to force widget updates
            if show and in_colab():
                clear_output(wait=True)
            for step in progress_bar(
                step_iter,
                (show_progress and not quiet) or (show and in_colab()),
                self._step_display
            ):
                if self._stop:
                    self.status = "stopped"
                    if interrupt:
                        raise KeyboardInterrupt()
                    break
                if function is not None:
                    if isinstance(function, (list, tuple)):
                        if len(function) < len(self._robots):
                            print_once("WARNING: you have not provided a controller function for every robot")
                        # Deterministically run robots round-robin:
                        stop_values = [
                            function[i](self._robots[i])
                            for i in range(len(function))
                            if function[i] is not None
                        ]
                    else:
                        stop_values = [function(self)]
                    stop = any(stop_values)
                    if stop:
                        break
                self._step(time_step, show=show, real_time=real_time)
                if callback is not None:
                    callback(self)
        self.status = "stopped"
        stop_real_time = time.monotonic()
        stop_time = self.time
        speed = (stop_time - start_time) / (stop_real_time - start_real_time)
        if steps > 1 and not quiet and not self.quiet:
            print(
                "Simulation stopped at: %s; speed %s x real time"
                % (format_time(self.time), round(speed, 2))
            )
        if show:
            self.draw()  # force to update any displays
        for value in stop_values:
            return value

    def _compute_complexity(self):
        # Proxy for how much drawing
        return sum([len(wall.lines) for wall in self._walls])

    def _step(self, time_step=None, show=True, real_time=True):
        """
        Run the simulator for 1 step.

        Args:
            time_step (float): the time unit to advance the simulation
            show (bool): if True, update the watchers
            real_time (bool): if True, run in real time, even introducing a
                delay if necessary
        """
        if time_step is not None and not isinstance(time_step, Number):
            raise ValueError(
                "Invalid time_step: %r; should be a number or None" % time_step
            )
        if not isinstance(show, bool):
            raise ValueError("Invalid show: %r; should be a bool" % show)
        if not isinstance(real_time, bool):
            raise ValueError("Invalid real_time: %r; should be a bool" % real_time)

        # Throttle needs to take into account the async update time
        # So as not to overwhelm the system. We give 0.1 time
        # per robot. This can be optimized to reduce the load.

        if self._backend.is_async():
            self._throttle_period = self._backend.get_dynamic_throttle(self)

        time_step = time_step if time_step is not None else self.time_step
        start_time = time.monotonic()
        for robot in self._robots:
            robot._step(time_step)
        self.time += time_step
        self.time = round(self.time, self._time_decimal_places)
        self.update(show)
        self._update_watchers()
        if show:
            now = time.monotonic()
            time_passed = now - start_time
            if real_time:  # real_time is ignored if not show
                sleep_time = self.time_step - time_passed
                # Tries to sleep enough to make even with real time/throttle time:
                # If running faster than real time/throttle period, need to wait some
                if sleep_time >= 0:
                    # Sleep even more for slow-motion:
                    time.sleep(sleep_time)
                # else it is already running slower than real time
            elif not self._backend.is_async():
                # Goal is to keep time_passed less than % of throttle period:
                if time_passed > self._throttle_period * self._show_throttle_percentage:
                    self._throttle_period += time_step

    def update(self, show=True):
        """
        Update the world, robots, and devices. Optionally, draw the
        world.

        Args:
            show (bool): if True, draw the world.
        """
        ## Update robots:
        self._draw_list = self._overlay_list[:]
        for robot in self._robots:
            robot.update(self._draw_list)
        if show:
            self._request_draw()

    def _request_draw(self):
        """
        Draw the world. This function is throttled
        """
        # Throttle:
        now = time.monotonic()
        time_since_last_call = now - self._time_of_last_call

        if time_since_last_call > self._throttle_period:
            self._time_of_last_call = now
            # End of throttle code

            self.draw()  # force

    def _get_robot_bulbs(self):
        bulbs = []
        for robot in self.robots:
            for device in robot._devices:
                if device.type == "bulb":
                    bulbs.append(device)
        return bulbs

    def _get_light_sources(self, all=False):
        if all:
            return [bulb for bulb in (self._bulbs + self._get_robot_bulbs()) if bulb.state == "on"]
        else:
            return [bulb for bulb in self._bulbs if bulb.state == "on"]

    def draw(self):
        """
        Force a redraw of the world.
        """
        if self._backend is None:
            return

        self._grid.update(self._food)

        with self._backend:
            self._backend.clear()
            self._backend.noStroke()
            if self._ground_image is not None:
                self._backend.draw_image(self._ground_image, 0, 0)
            else:
                self._backend.set_fill(self.ground_color)
                self._backend.draw_rect(0, 0, self.width, self.height)

            if len(self._food) > 0:
                smell = self._grid.get_image()
                smell = smell.resize((int(self.width * self.scale), int(self.height * self.scale)))
                self._backend.image.paste(smell, (0, 0), smell)

            ## Draw all bulbs in world:
            maxRange = max(self.width, self.height)
            for bulb in self._get_light_sources(all=True):
                if bulb.state == "on":
                    color = bulb.color
                    self._backend.line_width = 0
                    self._backend.noStroke()
                    x, y = bulb.get_position(world=True)
                    # Cast rays once:
                    all_hits = []
                    for ray in range(bulb.rays):
                        angle = ray/bulb.rays * TWO_PI
                        hits = cast_ray(self, bulb.robot, x, y, -angle + PI_OVER_2, maxRange)
                        all_hits.append(hits)
                    # Now draw the rings:
                    for i in range(bulb.draw_rings):
                        radius = (bulb.draw_rings - i) * 2
                        ray_length = bulb.brightness/30 * radius
                        color.alpha = (i + 1)/bulb.draw_rings * 255
                        self._backend.set_fill_style(color)
                        points = []
                        for ray in range(bulb.rays):
                            hits = all_hits[ray]
                            ray_length = bulb.brightness/30 * radius
                            if len(hits) > 0:
                                if hits[-1].distance < ray_length:
                                    ray_length = hits[-1].distance
                            angle = ray/bulb.rays * TWO_PI
                            x2, y2 = rotate_around(x, y, ray_length, angle)
                            points.append((x2, y2))
                        self._backend.draw_polygon(points)
                    self._backend.line_width = 1

            ## Draw walls:
            for wall in self._walls:
                if wall.wtype == "wall":
                    c = wall.color
                    if len(wall.lines) == 1:
                        self._backend.strokeStyle(c, 5)
                        self._backend.draw_line(
                            wall.lines[0].p1.x,
                            wall.lines[0].p1.y,
                            wall.lines[0].p2.x,
                            wall.lines[0].p2.y,
                        )
                    else:
                        self._backend.set_fill(c)
                        self._backend.noStroke()
                        self._backend.beginShape()
                        for line in wall.lines:
                            self._backend.vertex(line.p1.x, line.p1.y)
                            self._backend.vertex(line.p2.x, line.p2.y)
                        self._backend.endShape()

                    self._backend.lineWidth(1)
                    self._backend.noStroke()

                elif wall.wtype == "boundary":
                    c = wall.color
                    self._backend.strokeStyle(c, 3)
                    self._backend.draw_line(
                        wall.lines[0].p1.x,
                        wall.lines[0].p1.y,
                        wall.lines[0].p2.x,
                        wall.lines[0].p2.y,
                    )
                    self._backend.lineWidth(1)
                    self._backend.noStroke()

            ## Draw robots:
            for robot in self._robots:
                robot.draw(self._backend)

            # Draw the beacon:
            if self.beacon is not None:
                if self.beacon.state == "on":
                    self._backend.set_fill_style(BLUE)
                else:
                    self._backend.set_fill_style(BLACK)
                x, y = self.beacon.get_position(world=True)
                self._backend.strokeStyle(BLACK, 1)
                self._backend.draw_circle(x, y, 5)

            text = format_time(self.time)
            self._backend.draw_status(text)

            for items in self._draw_list:
                if len(items) == 1:
                    command = items[0]
                    args = tuple()
                    kwargs = {}
                elif len(items) == 2:
                    command = items[0]
                    args = items[1]
                    kwargs = {}
                elif len(items) == 3:
                    command = items[0]
                    args = items[1]
                    kwargs = items[2]

                self._backend.do_command(command, *args, **kwargs)

        self._draw_watchers()



"""
Handles creation of genomes, either from scratch or by sexual or
asexual reproduction from parents.
"""
from __future__ import division

import math
import random
from itertools import count

from neat.config import ConfigParameter, DefaultClassConfig
from neat.math_util import mean

# TODO: Provide some sort of optional cross-species performance criteria, which
# are then used to control stagnation and possibly the mutation rate
# configuration. This scheme should be adaptive so that species do not evolve
# to become "cautious" and only make very slow progress.


class DefaultReproduction(DefaultClassConfig):
    """
    Implements the default NEAT-python reproduction scheme:
    explicit fitness sharing with fixed-time species stagnation.
    """

    @classmethod
    def parse_config(cls, param_dict):
        return DefaultClassConfig(param_dict,
                                  [ConfigParameter('elitism', int, 0),
                                   ConfigParameter('survival_threshold', float, 0.2),
                                   ConfigParameter('min_species_size', int, 2)])

    def __init__(self, config, reporters, stagnation):
        # pylint: disable=super-init-not-called
        self.reproduction_config = config
        self.reporters = reporters
        self.genome_indexer = count(1)
        self.stagnation = stagnation
        self.ancestors = {}

    def create_new(self, genome_type, genome_config, num_genomes):
        new_genomes = {}
        for i in range(num_genomes):
            key = next(self.genome_indexer)
            g = genome_type(key)
            g.configure_new(genome_config)
            new_genomes[key] = g
            self.ancestors[key] = tuple()

        return new_genomes

    @staticmethod
    def compute_spawn(adjusted_fitness, previous_sizes, pop_size, min_species_size):
        """Compute the proper number of offspring per species (proportional to fitness)."""
        af_sum = sum(adjusted_fitness)

        spawn_amounts = []
        for af, ps in zip(adjusted_fitness, previous_sizes):
            if af_sum > 0:
                s = max(min_species_size, af / af_sum * pop_size)
            else:
                s = min_species_size

            d = (s - ps) * 0.5
            c = int(round(d))
            spawn = ps
            if abs(c) > 0:
                spawn += c
            elif d > 0:
                spawn += 1
            elif d < 0:
                spawn -= 1

            spawn_amounts.append(spawn)

        # Normalize the spawn amounts so that the next generation is roughly
        # the population size requested by the user.
        total_spawn = sum(spawn_amounts)
        norm = pop_size / total_spawn
        spawn_amounts = [max(min_species_size, int(round(n * norm))) for n in spawn_amounts]

        return spawn_amounts

    def reproduce(self, config, species, pop_size, generation):
        """
        Handles creation of genomes, either from scratch or by sexual or
        asexual reproduction from parents.
        """
        # TODO: I don't like this modification of the species and stagnation objects,
        # because it requires internal knowledge of the objects.

        # Filter out stagnated species, collect the set of non-stagnated
        # species members, and compute their average adjusted fitness.
        # The average adjusted fitness scheme (normalized to the interval
        # [0, 1]) allows the use of negative fitness values without
        # interfering with the shared fitness scheme.
        all_fitnesses = []
        remaining_species = []
        for stag_sid, stag_s, stagnant in self.stagnation.update(species, generation):
            if stagnant:
                self.reporters.species_stagnant(stag_sid, stag_s)
            else:
                all_fitnesses.extend(m.fitness for m in stag_s.members.values())
                remaining_species.append(stag_s)
        # The above comment was not quite what was happening - now getting fitnesses
        # only from members of non-stagnated species.

        # No species left.
        if not remaining_species:
            species.species = {}
            return {} # was []

        # Find minimum/maximum fitness across the entire population, for use in
        # species adjusted fitness computation.
        min_fitness = min(all_fitnesses)
        max_fitness = max(all_fitnesses)
        # Do not allow the fitness range to be zero, as we divide by it below.
        # TODO: The ``1.0`` below is rather arbitrary, and should be configurable.
        fitness_range = max(1.0, max_fitness - min_fitness)
        for afs in remaining_species:
            # Compute adjusted fitness.
            msf = mean([m.fitness for m in afs.members.values()])
            af = (msf - min_fitness) / fitness_range
            afs.adjusted_fitness = af

        adjusted_fitnesses = [s.adjusted_fitness for s in remaining_species]
        avg_adjusted_fitness = mean(adjusted_fitnesses) # type: float
        self.reporters.info("Average adjusted fitness: {:.3f}".format(avg_adjusted_fitness))

        # Compute the number of new members for each species in the new generation.
        previous_sizes = [len(s.members) for s in remaining_species]
        min_species_size = self.reproduction_config.min_species_size
        # Isn't the effective min_species_size going to be max(min_species_size,
        # self.reproduction_config.elitism)? That would probably produce more accurate tracking
        # of population sizes and relative fitnesses... doing. TODO: document.
        min_species_size = max(min_species_size, self.reproduction_config.elitism)
        spawn_amounts = self.compute_spawn(adjusted_fitnesses, previous_sizes,
                                           pop_size, min_species_size)

        new_population = {}
        species.species = {}
        for spawn, s in zip(spawn_amounts, remaining_species):
            # If elitism is enabled, each species always at least gets to retain its elites.
            spawn = max(spawn, self.reproduction_config.elitism)

            assert spawn > 0

            # The species has at least one member for the next generation, so retain it.
            old_members = list(s.members.items())
            s.members = {}
            species.species[s.key] = s

            # Sort members in order of descending fitness.
            old_members.sort(reverse=True, key=lambda x: x[1].fitness)

            # Transfer elites to new generation.
            if self.reproduction_config.elitism > 0:
                for i, m in old_members[:self.reproduction_config.elitism]:
                    new_population[i] = m
                    spawn -= 1

            if spawn <= 0:
                continue

            # Only use the survival threshold fraction to use as parents for the next generation.
            repro_cutoff = int(math.ceil(self.reproduction_config.survival_threshold *
                                         len(old_members)))
            # Use at least two parents no matter what the threshold fraction result is.
            repro_cutoff = max(repro_cutoff, 2)
            old_members = old_members[:repro_cutoff]

            # Randomly choose parents and produce the number of offspring allotted to the species.
            while spawn > 0:
                spawn -= 1

                parent1_id, parent1 = random.choice(old_members)
                parent2_id, parent2 = random.choice(old_members)

                # Note that if the parents are not distinct, crossover will produce a
                # genetically identical clone of the parent (but with a different ID).
                gid = next(self.genome_indexer)
                child = config.genome_type(gid)
                child.configure_crossover(parent1, parent2, config.genome_config)
                child.mutate(config.genome_config)
                new_population[gid] = child
                self.ancestors[gid] = (parent1_id, parent2_id)

        return new_population



# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python AI utils
#
# Copyright (c) 2020 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import numpy as np
import PIL.Image
import os

from .utils import get_file

_filename = get_file(
    "flowers-100.zip",
    "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/flowers/flowers-100.zip",
    extract=True,
)

DATA_DIR = os.path.splitext(_filename)[0]

def get():
    data = []
    for i in range(100):
        image = PIL.Image.open(os.path.join(DATA_DIR, "Image_%s.jpg" % i))
        data.append(np.array(image))
    return data



"""Uses `pickle` to save and restore populations (and other aspects of the simulation state)."""
from __future__ import print_function

import gzip
import random
import time

try:
    import cPickle as pickle  # pylint: disable=import-error
except ImportError:
    import pickle  # pylint: disable=import-error

from neat.population import Population
from neat.reporting import BaseReporter


class Checkpointer(BaseReporter):
    """
    A reporter class that performs checkpointing using `pickle`
    to save and restore populations (and other aspects of the simulation state).
    """

    def __init__(self, generation_interval=100, time_interval_seconds=300,
                 filename_prefix='neat-checkpoint-'):
        """
        Saves the current state (at the end of a generation) every ``generation_interval`` generations or
        ``time_interval_seconds``, whichever happens first.

        :param generation_interval: If not None, maximum number of generations between save intervals
        :type generation_interval: int or None
        :param time_interval_seconds: If not None, maximum number of seconds between checkpoint attempts
        :type time_interval_seconds: float or None
        :param str filename_prefix: Prefix for the filename (the end will be the generation number)
        """
        self.generation_interval = generation_interval
        self.time_interval_seconds = time_interval_seconds
        self.filename_prefix = filename_prefix

        self.current_generation = None
        self.last_generation_checkpoint = -1
        self.last_time_checkpoint = time.time()

    def start_generation(self, generation):
        self.current_generation = generation

    def post_evaluate(self, config, population, species_set, best_genome):
        checkpoint_due = False

        if self.time_interval_seconds is not None:
            dt = time.time() - self.last_time_checkpoint
            if dt >= self.time_interval_seconds:
                checkpoint_due = True

        if (checkpoint_due is False) and (self.generation_interval is not None):
            dg = self.current_generation - self.last_generation_checkpoint
            if dg >= self.generation_interval:
                checkpoint_due = True

        if checkpoint_due:
            self.save_checkpoint(config, population, species_set, best_genome, self.current_generation)
            self.last_generation_checkpoint = self.current_generation
            self.last_time_checkpoint = time.time()

    def save_checkpoint(self, config, population, species_set, best_genome, generation):
        """ Save the current simulation state. """
        filename = '{0}{1}'.format(self.filename_prefix, generation)
        print("Saving checkpoint to {0}".format(filename))

        with gzip.open(filename, 'w', compresslevel=5) as f:
            data = (generation, config, population, species_set, best_genome, random.getstate())
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def restore_checkpoint(filename):
        """Resumes the simulation from a previous saved point."""
        with gzip.open(filename) as f:
            generation, config, population, species_set, best_genome, rndstate = pickle.load(f)
            random.setstate(rndstate)
            return Population(config, (population, species_set, best_genome, generation))



# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python utils for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import math

try:
    from ipywidgets import (
        Button,
        Layout,
        GridBox,
        ButtonStyle,
        FloatSlider,
        HBox,
        VBox,
        GridspecLayout,
        TwoByTwoLayout,
    )
    _has_ipywidgets = True
except ImportError:
    _has_ipywidgets = False
    
try:
    from ipycanvas import Canvas, hold_canvas
    _has_ipycanvas = True
except ImportError:
    _has_ipycanvas = False


def has_ipycanvas():
    return _has_ipycanvas

def has_ipywidgets():
    return _has_ipywidgets

class Joystick():
    def __init__(self, scale=(1.0, 1.0), width=250, height=250, function=print):
        """
        A dynamic joystick drawn on a Canvas.

        Args:
            scale: the scaling of translate and rotate values; defaults (1.0, 1.0)
            width: width of the widget; default 250
            height: height of the widget; default 250
            function: the function to call when changing joystick; default print
        """
        self.translate_scale = scale[0]
        self.rotate_scale = scale[1]
        self.last_translate = 0
        self.last_rotate = 0
        self.state = "up"
        self.width = width
        self.height = height
        self.function = function
        self.canvas = Canvas(width=self.width, height=self.height)
        self.canvas.on_mouse_move(self.handle_mouse_move)
        self.canvas.on_mouse_down(self.handle_mouse_down)
        self.canvas.on_mouse_up(self.handle_mouse_up)
        self.canvas.layout.min_width = "%spx" % self.width
        self.canvas.layout.max_height = "%spx" % self.height
        # Draw blank joystick:
        self.reset()

    def clear(self):
        self.canvas.fill_style = "#B0C4DE"
        self.canvas.fill_rect(0, 0, self.width, self.height)
        self.canvas.stroke_style = "black"
        self.canvas.stroke_rect(0, 0, self.width, self.height)

    def rotate_around(self, x, y, length, angle):
        return (x + length * math.cos(-angle),
                y - length * math.sin(-angle))

    def handle_mouse_move(self, x, y):
        if self.state == "down":
            self.last_translate = ((self.height/2 - y) / (self.height/2)) * self.translate_scale
            self.last_rotate = ((self.width/2 - x) / (self.width/2)) * self.rotate_scale
            self.function(self.last_translate, self.last_rotate)
            with hold_canvas(self.canvas):
                self.clear()
                self.canvas.stroke_style = "black"
                angle = math.atan2(self.width/2 - x, self.height/2 - y)
                x1, y1 = self.rotate_around(self.width/2, self.height/2, 10, -angle)
                x2, y2 = self.rotate_around(self.width/2, self.height/2, -10, -angle)
                points = [
                    (self.width/2, self.height/2),
                    (x1, y1),
                    (x, y),
                    (x2, y2),
                ]
                self.canvas.fill_style = "gray"
                self.canvas.fill_polygon(points)
                self.canvas.fill_circle(self.width/2, self.height/2, 10)
                self.canvas.fill_style = "black"
                self.canvas.fill_circle(x, y, self.width/10)

    def handle_mouse_down(self, x, y):
        self.state = "down"
        self.handle_mouse_move(x, y)

    def handle_mouse_up(self, x, y):
        self.state = "up"
        self.last_translate = 0
        self.last_rotate = 0
        self.function(0, 0)
        self.reset()

    def reset(self):
        # translate, rotate
        self.last_translate = 0
        self.last_rotate = 0
        self.function(0, 0)
        self.clear()
        self.canvas.fill_style = "black"
        self.canvas.fill_circle(self.width/2, self.height/2, self.width/10)

    def watch(self):
        display(self.canvas)

    def get_widget(self):
        return self.canvas


class NoJoystick():
    def __init__(self, scale=(1.0, 1.0), width=250, height=250, function=print):
        """
        A simple joystick via buttons and sliders.

        Args:
            scale: the scaling of translate and rotate values; defaults (1.0, 1.0)
            function: the function to call when changing joystick; default print
            width: width of the widget (ignored)
            height: height of the widget (ignored)
        """
        self.translate_scale = scale[0]
        self.rotate_scale = scale[1]
        self.last_translate = 0
        self.last_rotate = 0

        self.function = function
        self.arrows = [
            "⬉ ⬆ ⬈",
            " ╲｜╱ ",
            "⬅－⊙－➡",
            " ╱｜╲ ",
            "⬋ ⬇ ⬊",
        ]

        self.movement = [
            [(1.0, -1.), (1.0, -.5), (1.0, 0.0), (1.0, 0.5), (1.0, 1.0)],
            [(0.5, -1.), (0.5, -.5), (0.5, 0.0), (0.5, 0.5), (0.5, 1.0)],
            [(0.0, -1.), (0.0, -.5), (0.0, 0.0), (0.0, 0.5), (0.0, 1.0)],
            [(-.5, -1.), (-.5, -.5), (-.5, 0.0), (-.5, 0.5), (-.5, 1.0)],
            [(-1., -1.), (-1., -.5), (-1., 0.0), (-1., 0.5), (-1., 1.0)],
        ]

        # Make the widgets:
        layout = Layout(height="35px", width="35px")
        self.buttons = []
        for row in range(5):
            for col in range(5):
                button = Button(description=self.arrows[row][col], layout=layout)
                button.on_click(self.create_move(row, col))
                self.buttons.append(button)

        self.rotate_slider = FloatSlider(
            min=-1, max=1, step=0.1,
            continuous_update=True,
            readout=False,
            layout=Layout(width="200px", height="30px"))

        self.translate_slider = FloatSlider(
            min=-1, max=1, step=0.1,
            orientation="vertical",
            continuous_update=True,
            readout=False,
            layout=Layout(width="30px"))

        self.array = GridBox(
            children=self.buttons,
            layout=Layout(
                grid_template_rows='40px 40px 40px 40px 40px',
                grid_template_columns='40px 40px 40px 40px 40px',
                grid_gap='0px 0px',
                overflow="intial",
            )
        )

        self.controls = TwoByTwoLayout(
            top_left=self.translate_slider,
            top_right=self.array,
            bottom_right=self.rotate_slider,
            width="min-content",
        )

        self.translate_slider.observe(self.on_translate_change,
                                      names='value')
        self.rotate_slider.observe(self.on_rotate_change,
                                   names='value')

    def on_translate_change(self, change):
        self.last_translate = change['new'] * self.translate_scale
        self.function(self.last_translate, None)

    def on_rotate_change(self, change):
        self.last_rotate = -change['new'] * self.rotate_scale
        self.function(None, self.last_rotate)

    def create_move(self, row, col):
        def on_click(button):
            translate, rotate = self.movement[row][col]
            self.translate_slider.value = translate
            self.rotate_slider.value = rotate
        return on_click

    def watch(self):
        display(self.controls)

    def get_widget(self):
        return self.controls



"""Threaded evaluation of genomes"""
from __future__ import print_function

import warnings

try:
    import threading
except ImportError: # pragma: no cover
    import dummy_threading as threading
    HAVE_THREADS = False
else:
    HAVE_THREADS = True

try:
    # pylint: disable=import-error
    import Queue as queue
except ImportError:
    # pylint: disable=import-error
    import queue


class ThreadedEvaluator(object):
    """
    A threaded genome evaluator.
    Useful on python implementations without GIL (Global Interpreter Lock).
    """
    def __init__(self, num_workers, eval_function):
        """
        eval_function should take two arguments (a genome object and the
        configuration) and return a single float (the genome's fitness).
        """
        self.num_workers = num_workers
        self.eval_function = eval_function
        self.workers = []
        self.working = False
        self.inqueue = queue.Queue()
        self.outqueue = queue.Queue()

        if not HAVE_THREADS: # pragma: no cover
            warnings.warn("No threads available; use ParallelEvaluator, not ThreadedEvaluator")

    def __del__(self):
        """
        Called on deletion of the object. We stop our workers here.
        WARNING: __del__ may not always work!
        Please stop the threads explicitly by calling self.stop()!
        TODO: ensure that there are no reference-cycles.
        """
        if self.working:
            self.stop()

    def start(self):
        """Starts the worker threads"""
        if self.working:
            return
        self.working = True
        for i in range(self.num_workers):
            w = threading.Thread(
                name="Worker Thread #{i}".format(i=i),
                target=self._worker,
                )
            w.daemon = True
            w.start()
            self.workers.append(w)

    def stop(self):
        """Stops the worker threads and waits for them to finish"""
        self.working = False
        for w in self.workers:
            w.join()
        self.workers = []

    def _worker(self):
        """The worker function"""
        while self.working:
            try:
                genome_id, genome, config = self.inqueue.get(
                    block=True,
                    timeout=0.2,
                    )
            except queue.Empty:
                continue
            f = self.eval_function(genome, config)
            self.outqueue.put((genome_id, genome, f))

    def evaluate(self, genomes, config):
        """Evaluate the genomes"""
        if not self.working:
            self.start()
        p = 0
        for genome_id, genome in genomes:
            p += 1
            self.inqueue.put((genome_id, genome, config))

        # assign the fitness back to each genome
        while p > 0:
            p -= 1
            ignored_genome_id, genome, fitness = self.outqueue.get()
            genome.fitness = fitness



"""Deals with the attributes (variable parameters) of genes"""
from random import choice, gauss, random, uniform
from neat.config import ConfigParameter


# TODO: There is probably a lot of room for simplification of these classes using metaprogramming.


class BaseAttribute(object):
    """Superclass for the type-specialized attribute subclasses, used by genes."""

    def __init__(self, name, **default_dict):
        self.name = name
        for n, default in default_dict.items():
            self._config_items[n] = [self._config_items[n][0], default]
        for n in self._config_items:
            setattr(self, n + "_name", self.config_item_name(n))

    def config_item_name(self, config_item_base_name):
        return "{0}_{1}".format(self.name, config_item_base_name)

    def get_config_params(self):
        return [ConfigParameter(self.config_item_name(n),
                                self._config_items[n][0],
                                self._config_items[n][1])
                for n in self._config_items]


class FloatAttribute(BaseAttribute):
    """
    Class for numeric attributes,
    such as the response of a node or the weight of a connection.
    """
    _config_items = {"init_mean": [float, None],
                     "init_stdev": [float, None],
                     "init_type": [str, 'gaussian'],
                     "replace_rate": [float, None],
                     "mutate_rate": [float, None],
                     "mutate_power": [float, None],
                     "max_value": [float, None],
                     "min_value": [float, None]}

    def clamp(self, value, config):
        min_value = getattr(config, self.min_value_name)
        max_value = getattr(config, self.max_value_name)
        return max(min(value, max_value), min_value)

    def init_value(self, config):
        mean = getattr(config, self.init_mean_name)
        stdev = getattr(config, self.init_stdev_name)
        init_type = getattr(config, self.init_type_name).lower()

        if ('gauss' in init_type) or ('normal' in init_type):
            return self.clamp(gauss(mean, stdev), config)

        if 'uniform' in init_type:
            min_value = max(getattr(config, self.min_value_name),
                            (mean - (2 * stdev)))
            max_value = min(getattr(config, self.max_value_name),
                            (mean + (2 * stdev)))
            return uniform(min_value, max_value)

        raise RuntimeError("Unknown init_type {!r} for {!s}".format(getattr(config,
                                                                            self.init_type_name),
                                                                    self.init_type_name))

    def mutate_value(self, value, config):
        # mutate_rate is usually no lower than replace_rate, and frequently higher -
        # so put first for efficiency
        mutate_rate = getattr(config, self.mutate_rate_name)

        r = random()
        if r < mutate_rate:
            mutate_power = getattr(config, self.mutate_power_name)
            return self.clamp(value + gauss(0.0, mutate_power), config)

        replace_rate = getattr(config, self.replace_rate_name)

        if r < replace_rate + mutate_rate:
            return self.init_value(config)

        return value

    def validate(self, config):  # pragma: no cover
        pass


class BoolAttribute(BaseAttribute):
    """Class for boolean attributes such as whether a connection is enabled or not."""
    _config_items = {"default": [str, None],
                     "mutate_rate": [float, None],
                     "rate_to_true_add": [float, 0.0],
                     "rate_to_false_add": [float, 0.0]}

    def init_value(self, config):
        default = str(getattr(config, self.default_name)).lower()

        if default in ('1', 'on', 'yes', 'true'):
            return True
        elif default in ('0', 'off', 'no', 'false'):
            return False
        elif default in ('random', 'none'):
            return bool(random() < 0.5)

        raise RuntimeError("Unknown default value {!r} for {!s}".format(default,
                                                                        self.name))

    def mutate_value(self, value, config):
        mutate_rate = getattr(config, self.mutate_rate_name)

        if value:
            mutate_rate += getattr(config, self.rate_to_false_add_name)
        else:
            mutate_rate += getattr(config, self.rate_to_true_add_name)

        if mutate_rate > 0:
            r = random()
            if r < mutate_rate:
                # NOTE: we choose a random value here so that the mutation rate has the
                # same exact meaning as the rates given for the string and bool
                # attributes (the mutation operation *may* change the value but is not
                # guaranteed to do so).
                return random() < 0.5

        return value

    def validate(self, config):  # pragma: no cover
        pass


class StringAttribute(BaseAttribute):
    """
    Class for string attributes such as the aggregation function of a node,
    which are selected from a list of options.
    """
    _config_items = {"default": [str, 'random'],
                     "options": [list, None],
                     "mutate_rate": [float, None]}

    def init_value(self, config):
        default = getattr(config, self.default_name)

        if default.lower() in ('none', 'random'):
            options = getattr(config, self.options_name)
            return choice(options)

        return default

    def mutate_value(self, value, config):
        mutate_rate = getattr(config, self.mutate_rate_name)

        if mutate_rate > 0:
            r = random()
            if r < mutate_rate:
                options = getattr(config, self.options_name)
                return choice(options)

        return value

    def validate(self, config):  # pragma: no cover
        pass



# -*- coding: utf-8 -*-
# ******************************************************
# aitk.networks: Keras model wrapper with visualizations
#
# Copyright (c) 2021 Douglas S. Blank
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.networks
#
# ******************************************************

import matplotlib.pyplot as plt
from tensorflow.keras.callbacks import Callback, EarlyStopping


def match_acc(name):
    return (name.endswith("acc") or
            name.endswith("accuracy"))

def match_loss(name):
    return name == "loss"

def match_val(name):
    return name.startswith("val_")

def make_early_stop(monitor, patience):
    return EarlyStopping(monitor=monitor, patience=patience, verbose=True)

def make_stop(metric, goal, patience, use_validation):
    return StopWhen(metric, goal, patience, use_validation)

def make_save(network, save_rate):
    return SaveWeights(network, save_rate)

class UpdateCallback(Callback):
    def __init__(self, network, report_rate):
        super().__init__()
        self._network = network
        self._report_rate = report_rate
        self._figure = None

    def on_train_begin(self, logs=None):
        print("Training %s..." % self._network.name)

    def on_epoch_end(self, epoch, logs=None):
        self._network.on_epoch_end(self, logs, self._report_rate)

    def on_train_end(self, logs=None):
        self._network.on_epoch_end(self, logs)
        if self._figure is not None:
            plt.close()

class SaveWeights(Callback):
    def __init__(self, network, save_rate):
        super().__init__()
        self.network = network
        self.save_rate = save_rate

    def on_train_begin(self, logs=None):
        """
        Save the initial weights
        """
        if len(self.network._history["weights"]) == 0:
            self.network._history["weights"].append((0, self.network.get_weights()))

    def on_epoch_end(self, epoch, logs=None):
        if (epoch + 1) % self.save_rate == 0:
            self.network._history["weights"].append((epoch + 1, self.network.get_weights()))

class StopWhen(Callback):
    def __init__(self, metric="acc", goal=1.0, patience=0, use_validation=False, verbose=True):
        super().__init__()
        self.metric = metric
        self.goal = goal
        self.patience = patience
        self.use_validation = use_validation
        self.wait = 0
        self.verbose = verbose

    def on_train_begin(self, logs=None):
        self.wait = 0
        self.i_stopped_it = False

    def on_train_end(self, logs=None):
        if self.i_stopped_it and self.verbose > 0:
            prefix = "val_" if self.use_validation else ""
            item = "%s%s" % (prefix, self.metric)
            print("Stopped because %s beat goal of %s" % (item, self.goal))

    def compare(self, value, goal):
        # self.metric is either "accuracy" or "loss"
        if self.metric == "accuracy":
            return value >= goal
        else:
            return value <= goal

    def on_epoch_end(self, epoch, logs=None):
        if logs is not None:
            metric_value = None
            for key in logs:
                if self.use_validation:
                    if self.metric == "accuracy" and match_acc(key) and match_val(key):
                        metric_value = logs[key]
                        break
                    elif self.metric == "loss" and match_loss(key) and match_val(key):
                        metric_value = logs[key]
                        break
                else:
                    if self.metric == "accuracy" and match_acc(key) and not match_val(key):
                        metric_value = logs[key]
                        break
                    elif self.metric == "loss" and match_loss(key) and not match_val(key):
                        metric_value = logs[key]
                        break
            if metric_value is not None:
                if self.compare(metric_value, self.goal):
                    if self.wait >= self.patience:
                        self.model.stop_training = True
                        self.i_stopped_it = True
                    else:
                        self.wait += 1
                else:
                    # else, go back to zero and start over
                    self.wait = 0



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import math

from ..utils import distance, rotate_around, normal_dist, uniform_angle
from .base import BaseDevice

from aitk.utils import Color


BLUE = Color("blue", 64)
BLACK = Color("black", 64)
WHITE = Color("white", 64)

class Compass(BaseDevice):
    def __init__(self, position=(0, 0), name="compass", **kwargs):
        """
        A compass for detecting direction to beacon.

        Args:
            position (Tuple[int, int]): the position of the
                sensor in centimeters relative to the center
                of the robot.
            name (str): the name of the device
        """
        config = {
            "position": position,
            "name": name,
        }
        config.update(kwargs)
        self._watcher = None
        self.robot = None
        self.initialize()
        self.from_json(config)

    def __repr__(self):
        return "<Compass %r position=%r>" % (self.name, self.position)

    def initialize(self):
        """
        Internal method to set all settings to default values.
        """
        self.type = "compass"
        self.name = "compass"
        self.value = 0.0
        self.values = [0, 0, 0, 0]
        self.length = 30  # CM
        self.position = [0, 0]
        self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
        self.dir_from_center = math.atan2(-self.position[0], self.position[1])

    def from_json(self, config):
        """
        Set the settings from a device config.

        Args:
            config (dict): a config dictionary
        """
        valid_keys = set([
            "position", "name", "class"
        ])
        self.verify_config(valid_keys, config)

        if "name" in config:
            self.name = config["name"]
        if "position" in config:
            self.position = config["position"]
            # Get location of sensor, doesn't change once position is set:
            self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
            self.dir_from_center = math.atan2(-self.position[0], self.position[1])

    def to_json(self):
        """
        Save the internal settings to a config dictionary.
        """
        config = {
            "class": self.__class__.__name__,
            "position": self.position,
            "name": self.name,
        }
        return config

    def _step(self, time_step):
        pass

    def update(self, draw_list=None):
        """
        Update the device.

        Args:
            draw_list (list): optional. If given, then the
                method can add to it for drawing later.
        """
        self.value = 0
        self.values = [0, 0, 0, 0]
        # Location of sensor:
        p = rotate_around(
            self.robot.x,
            self.robot.y,
            self.dist_from_center,
            self.robot.a + self.dir_from_center + math.pi / 2,
        )
        beacon = self.robot.world.beacon
        if beacon is not None and beacon.state == "on":
            x, y = beacon.get_position(world=True)
            to_beacon = uniform_angle(math.pi * 3/2 - math.atan2(p[0] - x, p[1] - y))
            front = uniform_angle(self.robot.a)
            self.value = uniform_angle(to_beacon - front)

            segment = 2 * math.pi / 4.0
            MAX = math.pi * 2
            # 0: front, 1: counter-clockwise, left, 2: back, 3: right
            if (0 < self.value < segment/2) or ((MAX - segment/2) < self.value < MAX):
                self.values[0] = 1
            elif (1 * segment/2 < self.value < 3 * segment/2):
                self.values[1] = 1
            elif (3 * segment/2 < self.value < 5 * segment/2):
                self.values[2] = 1
            else:
                self.values[3] = 1

    def draw(self, backend):
        """
        Draw the device on the backend.

        Args:
            backend (Backend): an aitk drawing backend
        """
        def get_arc(quadrant):
            if quadrant == 0: # front
                start = - 1 * math.pi / 4
                stop = math.pi / 4
            elif quadrant == 1: # left
                start = 1 * math.pi / 4
                stop = 3 * math.pi / 4
            elif quadrant == 2: # back
                start = -3 * math.pi / 4
                stop = -5 * math.pi / 4
            elif quadrant == 3: # right
                start =  - 3 * math.pi / 4
                stop = - 1 * math.pi / 4
            return start, stop

        for quadrant in range(4):
            start, stop = get_arc(quadrant)
            if self.values[quadrant]:
                backend.lineWidth(1)
                backend.set_stroke_style(BLACK)
                backend.set_fill_style(BLUE)
            else:
                backend.lineWidth(1)
                backend.set_stroke_style(BLACK)
                backend.set_fill_style(WHITE)
            backend.draw_arc(0, 0, self.length, self.length, start, stop)

    def watch(self, title="Compass:"):
        """
        Create a dynamically updating view
        of this sensor.

        Args:
            title (str): title of sensor
        """
        widget = self.get_widget(title=title)
        return display(widget)

    def get_value(self):
        """
        Get the direction to the beacon. The value
        is in radians, with 0 pointing straight ahead of the
        robot, and increasing counter-clockwise back
        to 2 pi.
        """
        return self.value

    def get_values(self):
        """
        Get the values of the compass as an array of zeros
        and a single 1 in the quadrant indicating where the beacon
        is relative to the robot.

        The indices of the array are as shown relative to the
        robot:

        \  0  /
         \   /
          \ /
        1     3
          / \
         /   \
        /  2  \

        where 0=front, 1=left, 2=back, and 3=right.

        So that the meaning of the array is:

        [front, left, back, right]
        """
        return self.values

    def get_widget(self, title="Compass:"):
        """
        Return the dynamically updating widget.

        Args:
            title (str): title of sensor
        """
        from ..watchers import AttributesWatcher

        if self.robot is None or self.robot.world is None:
            print("ERROR: can't watch until added to robot, and robot is in world")
            return None

        if self._watcher is None:
            self._watcher = AttributesWatcher(
                self, "name", "value", "values", title=title, labels=["Name:", "Compass:", "Values:"]
            )
            self.robot.world._watchers.append(self._watcher)

        return self._watcher.widget

    def set_position(self, position):
        """
        Set the position of the compass with respect to the center of the
        robot.

        Args:
            position (List[int, int]): represents [x, y] in CM from
                center of robot
        """
        if len(position) != 2:
            raise ValueError("position must be of length two")

        self.position = position
        # Get location of sensor, doesn't change once position is set:
        self.dist_from_center = distance(0, 0, self.position[0], self.position[1])
        self.dir_from_center = math.atan2(-self.position[0], self.position[1])



# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python AI utils
#
# Copyright (c) 2020 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import os
import numpy as np
import glob
from PIL import Image

from .utils import get_file

_filename = get_file(
    "dogs-vs-cats-100.npz",
    "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/dogs-vs-cats/dogs-vs-cats-100.npz",
    extract=False,
)

def get_dogs_vs_cats_100():
    dataset = np.load(_filename)
    return dataset



# -*- coding: utf-8 -*-
# ****************************************************************
# aitk.algorithms: Algorithms for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.algorithms
#
# ****************************************************************

import random
import numpy as np
import matplotlib.pyplot as plt

class NoveltySearch():
    """
    Novelty search is a method of searching for a variety of solutions
    rather than trying to optimize a particular goal. To do this, novelty
    search maintains an archive of behaviors, comparing current behaviors
    against the saved ones to determine whether a current behavior is novel
    enough to add to the archive. A current behavior's novelty is defined
    as the average distance to its k nearest neighbors in the archive.
    Behaviors are represented as lists of points.
    """
    def __init__(self, k, limit, threshold, max_dist):
        self.archive = []              # archive of past unique behaviors
        self.k = k                     # k nearest neighbors determine novelty
        self.limit = limit             # maximum size of the archive
        self.threshold = threshold     # how distant to be considered novel
        self.max_dist = max_dist       # max dist possible between two behaviors
        self.growth = []               # size of the archive over time

    def point_distance(self, p1, p2):
        """
        Returns the Euclidean distance between points p1 and p2, which are
        assumed to be lists of equal length. This method works for points of
        any length.
        """
        assert len(p1) == len(p2), "Lengths of points must be the same"
        a1 = np.array(p1)
        a2 = np.array(p2)
        return np.sqrt(np.sum( (a1-a2)**2 ))

    def behavior_distance(self, b1, b2):
        """
        Returns the sum of the point distances between behaviors b1 and b2,
        which are assumed to be lists of points of equal length. This method
        works for behaviors of any length.
        """
        assert len(b1) == len(b2), "Lengths of behaviors must the same"
        return sum([self.point_distance(p,q) for p, q in zip(b1, b2)])

    def k_nearest_distance(self, behavior):
        """
        For the given behavior, returns the sum of the behavior distances to
        its k nearest neighbors in the archive. If there are less than k
        behaviors in the archive, then returns the sum of all of the behaviors.
        """
        dists = [self.behavior_distance(behavior, novel_b[0])
                 for novel_b in self.archive]
        dists.sort()
        return sum(dists[:self.k])

    def sparseness(self, behavior):
        """
        For the given behavior, returns the normalized average distance
        (based on the max_dist) to its k nearest neighbors in the archive.
        If the archive is empty, then return 1.0. If the archive has less
        than k behaviors, the use the archive's current size to compute
        the average distance. The sparseness of a behavior will represent
        its novelty.
        """
        if len(self.archive) == 0:
            return 1.0
        d = self.k_nearest_distance(behavior)
        avg = d / min(self.k, len(self.archive))
        return avg / self.max_dist

    def check_archive(self, behavior, other_info=None):
        """
        Checks whether the given behavior is novel enough to add to the
        archive and returns its sparseness. All new behaviors should be added
        to the end of the archive. The behavior at the front of the archive
        is the oldest. When the size of the archive is less than k, then
        add the given behavior to the archive. Otherwise check if the
        sparseness of the given behavior is greater than the threshold, and
        if so, add the behavior to the archive. After adding a behavior, test
        if the archive size has exceeded the limit, and if so, remove the
        oldest behavior from the archive.
        """
        novelty = self.sparseness(behavior)
        if novelty > self.threshold or len(self.archive) < self.k:
            self.archive.append([behavior, novelty, other_info])
        if len(self.archive) > self.limit:
            self.archive.pop(0)
        self.growth.append(len(self.archive))
        return novelty

    def plot_growth(self):
        """
        Plot growth or archive size over time.
        """
        assert len(self.growth) > 0, "archive must be non-empty to plot growth"
        plt.title("Archive Growth")
        plt.xlabel("Time")
        plt.ylabel("Size")
        plt.plot(self.growth)
        plt.show()




# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python AI utils
#
# Copyright (c) 2020 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import os
import numpy as np

from .utils import get_file

_filename = get_file(
    "digits6x6.zip",
    "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/digits6x6/digits6x6.zip",
    extract=True,
)

DATA_DIR = os.path.splitext(_filename)[0]

def onehot(num):
    retval = [0] * 10
    retval[num] = 1
    return retval

def get():
    """Each digit is represented by a 6x6 grid of 1's and 0's separated by blank lines."""
    filename = DATA_DIR + ".data"
    fp = open(filename, "r")
    data = []
    while True:
        digit = []
        line = fp.readline()
        while len(line) != 0 and (line[0] == '0' or line[0] == '1'):
            line.strip()
            values = line.split()
            digit += [int(v) for v in values]
            line = fp.readline()
        if len(line) == 0:
            break
        if len(digit) > 0:
            data.append(digit)
    array = np.array(data)
    inputs = array.reshape((240, 6, 6))
    targets = np.array([onehot(v) for v in list(range(10)) * 24])
    return inputs, targets



"""Handles genomes (individuals in the population)."""
from __future__ import division, print_function


from itertools import count
from random import choice, random, shuffle

import sys

from neat.activations import ActivationFunctionSet
from neat.aggregations import AggregationFunctionSet
from neat.config import ConfigParameter, write_pretty_params
from neat.genes import DefaultConnectionGene, DefaultNodeGene
from neat.graphs import creates_cycle


class DefaultGenomeConfig(object):
    """Sets up and holds configuration information for the DefaultGenome class."""
    allowed_connectivity = ['unconnected', 'fs_neat_nohidden', 'fs_neat', 'fs_neat_hidden',
                            'full_nodirect', 'full', 'full_direct',
                            'partial_nodirect', 'partial', 'partial_direct']
    nw_types = ['default', 'rnn', 'mlrnn', 'ctrnn', 'iznn']

    def __init__(self, params):
        # Create full set of available activation functions.
        self.activation_defs = ActivationFunctionSet()
        # ditto for aggregation functions - name difference for backward compatibility
        self.aggregation_function_defs = AggregationFunctionSet()
        self.aggregation_defs = self.aggregation_function_defs

        self._params = [ConfigParameter('num_inputs', int),
                        ConfigParameter('num_outputs', int),
                        ConfigParameter('num_hidden', int),
                        ConfigParameter('nw_type', str, default='default'),
                        ConfigParameter('feed_forward', bool),
                        ConfigParameter('compatibility_disjoint_coefficient', float),
                        ConfigParameter('compatibility_weight_coefficient', float),
                        ConfigParameter('conn_add_prob', float),
                        ConfigParameter('conn_delete_prob', float),
                        ConfigParameter('node_add_prob', float),
                        ConfigParameter('node_delete_prob', float),
                        ConfigParameter('recurrent_con_prob', float, default=0.5),
                        ConfigParameter('single_structural_mutation', bool, 'false'),
                        ConfigParameter('structural_mutation_surer', str, 'default'),
                        ConfigParameter('initial_connection', str, 'unconnected')]

        # Gather configuration data from the gene classes.
        self.node_gene_type = params['node_gene_type']
        self._params += self.node_gene_type.get_config_params()
        self.connection_gene_type = params['connection_gene_type']
        self._params += self.connection_gene_type.get_config_params()

        # Use the configuration data to interpret the supplied parameters.
        for p in self._params:
            setattr(self, p.name, p.interpret(params))

        # By convention, input pins have negative keys, and the output
        # pins have keys 0,1,...
        self.input_keys = [-i - 1 for i in range(self.num_inputs)]
        self.output_keys = [i for i in range(self.num_outputs)]

        assert self.nw_type in self.nw_types

        self.connection_fraction = None

        # Verify that initial connection type is valid.
        # pylint: disable=access-member-before-definition
        if 'partial' in self.initial_connection:
            c, p = self.initial_connection.split()
            self.initial_connection = c
            self.connection_fraction = float(p)
            if not (0 <= self.connection_fraction <= 1):
                raise RuntimeError(
                    "'partial' connection value must be between 0.0 and 1.0, inclusive.")

        assert self.initial_connection in self.allowed_connectivity

        # Verify structural_mutation_surer is valid.
        # pylint: disable=access-member-before-definition
        if self.structural_mutation_surer.lower() in ['1', 'yes', 'true', 'on']:
            self.structural_mutation_surer = 'true'
        elif self.structural_mutation_surer.lower() in ['0', 'no', 'false', 'off']:
            self.structural_mutation_surer = 'false'
        elif self.structural_mutation_surer.lower() == 'default':
            self.structural_mutation_surer = 'default'
        else:
            error_string = "Invalid structural_mutation_surer {!r}".format(
                self.structural_mutation_surer)
            raise RuntimeError(error_string)

        self.node_indexer = None

    def add_activation(self, name, func):
        self.activation_defs.add(name, func)

    def add_aggregation(self, name, func):
        self.aggregation_function_defs.add(name, func)

    def save(self, f):
        if 'partial' in self.initial_connection:
            if not (0 <= self.connection_fraction <= 1):
                raise RuntimeError(
                    "'partial' connection value must be between 0.0 and 1.0, inclusive.")
            f.write('initial_connection      = {0} {1}\n'.format(self.initial_connection,
                                                                 self.connection_fraction))
        else:
            f.write('initial_connection      = {0}\n'.format(self.initial_connection))

        assert self.initial_connection in self.allowed_connectivity

        write_pretty_params(f, self, [p for p in self._params
                                      if 'initial_connection' not in p.name])

    def get_new_node_key(self, node_dict):
        if self.node_indexer is None:
            self.node_indexer = count(max(list(node_dict)) + 1)

        new_id = next(self.node_indexer)

        assert new_id not in node_dict

        return new_id

    def check_structural_mutation_surer(self):
        if self.structural_mutation_surer == 'true':
            return True
        elif self.structural_mutation_surer == 'false':
            return False
        elif self.structural_mutation_surer == 'default':
            return self.single_structural_mutation
        else:
            error_string = "Invalid structural_mutation_surer {!r}".format(
                self.structural_mutation_surer)
            raise RuntimeError(error_string)


class DefaultGenome(object):
    """
    A genome for generalized neural networks.

    Terminology
        pin: Point at which the network is conceptually connected to the external world;
             pins are either input or output.
        node: Analog of a physical neuron.
        connection: Connection between a pin/node output and a node's input, or between a node's
             output and a pin/node input.
        key: Identifier for an object, unique within the set of similar objects.

    Design assumptions and conventions.
        1. Each output pin is connected only to the output of its own unique
           neuron by an implicit connection with weight one. This connection
           is permanently enabled.
        2. The output pin's key is always the same as the key for its
           associated neuron.
        3. Output neurons can be modified but not deleted.
        4. The input values are applied to the input pins unmodified.
    """

    @classmethod
    def parse_config(cls, param_dict):
        param_dict['node_gene_type'] = DefaultNodeGene
        param_dict['connection_gene_type'] = DefaultConnectionGene
        return DefaultGenomeConfig(param_dict)

    @classmethod
    def write_config(cls, f, config):
        config.save(f)

    def __init__(self, key):
        # Unique identifier for a genome instance.
        self.key = key

        # (gene_key, gene) pairs for gene sets.
        self.connections = {}
        self.nodes = {}

        # Fitness results.
        self.fitness = None

    def configure_new(self, config):
        """Configure a new genome based on the given configuration."""

        # Create node genes for the output pins.
        for node_key in config.output_keys:
            self.nodes[node_key] = self.create_node(config, node_key)

        # Add hidden nodes if requested.
        if config.num_hidden > 0:
            for i in range(config.num_hidden):
                node_key = config.get_new_node_key(self.nodes)
                assert node_key not in self.nodes
                node = self.create_node(config, node_key)
                self.nodes[node_key] = node

        # Add connections based on initial connectivity type.

        if 'fs_neat' in config.initial_connection:
            if config.initial_connection == 'fs_neat_nohidden':
                self.connect_fs_neat_nohidden(config)
            elif config.initial_connection == 'fs_neat_hidden':
                self.connect_fs_neat_hidden(config)
            else:
                if config.num_hidden > 0:
                    print(
                        "Warning: initial_connection = fs_neat will not connect to hidden nodes;",
                        "\tif this is desired, set initial_connection = fs_neat_nohidden;",
                        "\tif not, set initial_connection = fs_neat_hidden",
                        sep='\n', file=sys.stderr)
                self.connect_fs_neat_nohidden(config)
        elif 'full' in config.initial_connection:
            if config.initial_connection == 'full_nodirect':
                self.connect_full_nodirect(config)
            elif config.initial_connection == 'full_direct':
                self.connect_full_direct(config)
            else:
                if config.num_hidden > 0:
                    print(
                        "Warning: initial_connection = full with hidden nodes will not do direct input-output connections;",
                        "\tif this is desired, set initial_connection = full_nodirect;",
                        "\tif not, set initial_connection = full_direct",
                        sep='\n', file=sys.stderr)
                self.connect_full_nodirect(config)
        elif 'partial' in config.initial_connection:
            if config.initial_connection == 'partial_nodirect':
                self.connect_partial_nodirect(config)
            elif config.initial_connection == 'partial_direct':
                self.connect_partial_direct(config)
            else:
                if config.num_hidden > 0:
                    print(
                        "Warning: initial_connection = partial with hidden nodes will not do direct input-output connections;",
                        "\tif this is desired, set initial_connection = partial_nodirect {0};".format(
                            config.connection_fraction),
                        "\tif not, set initial_connection = partial_direct {0}".format(
                            config.connection_fraction),
                        sep='\n', file=sys.stderr)
                self.connect_partial_nodirect(config)

    def configure_crossover(self, genome1, genome2, config):
        """ Configure a new genome by crossover from two parent genomes. """
        if genome1.fitness > genome2.fitness:
            parent1, parent2 = genome1, genome2
        else:
            parent1, parent2 = genome2, genome1

        # Inherit connection genes
        for key, cg1 in parent1.connections.items():
            cg2 = parent2.connections.get(key)
            if cg2 is None:
                # Excess or disjoint gene: copy from the fittest parent.
                self.connections[key] = cg1.copy()
            else:
                # Homologous gene: combine genes from both parents.
                self.connections[key] = cg1.crossover(cg2)

        # Inherit node genes
        parent1_set = parent1.nodes
        parent2_set = parent2.nodes

        for key, ng1 in parent1_set.items():
            ng2 = parent2_set.get(key)
            assert key not in self.nodes
            if ng2 is None:
                # Extra gene: copy from the fittest parent
                self.nodes[key] = ng1.copy()
            else:
                # Homologous gene: combine genes from both parents.
                self.nodes[key] = ng1.crossover(ng2)

    def mutate(self, config):
        """ Mutates this genome. """

        if config.single_structural_mutation:
            div = max(1, (config.node_add_prob + config.node_delete_prob +
                          config.conn_add_prob + config.conn_delete_prob))
            r = random()
            if r < (config.node_add_prob/div):
                self.mutate_add_node(config)
            elif r < ((config.node_add_prob + config.node_delete_prob)/div):
                self.mutate_delete_node(config)
            elif r < ((config.node_add_prob + config.node_delete_prob +
                       config.conn_add_prob)/div):
                self.mutate_add_connection(config)
            elif r < ((config.node_add_prob + config.node_delete_prob +
                       config.conn_add_prob + config.conn_delete_prob)/div):
                self.mutate_delete_connection()
        else:
            if random() < config.node_add_prob:
                self.mutate_add_node(config)

            if random() < config.node_delete_prob:
                self.mutate_delete_node(config)

            if random() < config.conn_add_prob:
                self.mutate_add_connection(config)

            if random() < config.conn_delete_prob:
                self.mutate_delete_connection()

        # Mutate connection genes.
        for cg in self.connections.values():
            cg.mutate(config)

        # Mutate node genes (bias, response, etc.).
        for ng in self.nodes.values():
            ng.mutate(config)

    def mutate_add_node(self, config):
        if not self.connections:
            if config.check_structural_mutation_surer():
                self.mutate_add_connection(config)
            return

        # Choose a random connection to split
        conn_to_split = choice(list(self.connections.values()))
        new_node_id = config.get_new_node_key(self.nodes)
        ng = self.create_node(config, new_node_id)
        self.nodes[new_node_id] = ng

        # Disable this connection and create two new connections joining its nodes via
        # the given node.  The new node+connections have roughly the same behavior as
        # the original connection (depending on the activation function of the new node).
        conn_to_split.enabled = False

        i, o = conn_to_split.key
        self.add_connection(config, i, new_node_id, 1.0, True)
        self.add_connection(config, new_node_id, o, conn_to_split.weight, True)

    def add_connection(self, config, input_key, output_key, weight, enabled):
        # TODO: Add further validation of this connection addition?
        assert isinstance(input_key, int)
        assert isinstance(output_key, int)
        assert output_key >= 0
        assert isinstance(enabled, bool)
        key = (input_key, output_key)
        connection = config.connection_gene_type(key)
        connection.init_attributes(config)
        connection.weight = weight
        connection.enabled = enabled
        self.connections[key] = connection

    def mutate_add_connection(self, config):
        """
        Attempt to add a new connection, the only restriction being that the output
        node cannot be one of the network input pins.
        """
        possible_outputs = list(self.nodes)
        out_node = choice(possible_outputs)

        possible_inputs = possible_outputs + config.input_keys
        in_node = choice(possible_inputs)

        # Don't duplicate connections.
        key = (in_node, out_node)
        if key in self.connections:
            # TODO: Should this be using mutation to/from rates? Hairy to configure...
            if config.check_structural_mutation_surer():
                self.connections[key].enabled = True
            return

        # Don't allow connections between two output nodes
        if in_node in config.output_keys and out_node in config.output_keys:
            return

        # No need to check for connections between input nodes:
        # they cannot be the output end of a connection (see above).

        # For feed-forward networks, avoid creating cycles.
        if config.feed_forward and creates_cycle(list(self.connections), key):
            return

        # If this is a MLRecurrentNetwork setup, there's a chance of a recurrent edge
        if in_node >= 0 and config.nw_type == 'mlrnn':
            if random() > config.recurrent_con_prob:
                in_node = -in_node - config.num_inputs - 1

        cg = self.create_connection(config, in_node, out_node)
        self.connections[cg.key] = cg

    def mutate_delete_node(self, config):
        # Do nothing if there are no non-output nodes.
        available_nodes = [k for k in self.nodes if k not in config.output_keys]
        if not available_nodes:
            return -1

        del_key = choice(available_nodes)

        connections_to_delete = set()
        for k, v in self.connections.items():
            if del_key in v.key:
                connections_to_delete.add(v.key)
            if config.nw_type == 'mlrnn':  # Check recurrent connections from our node
                if -del_key - config.num_inputs - 1 == v.key[0]:
                    connections_to_delete.add(v.key)

        for key in connections_to_delete:
            del self.connections[key]

        del self.nodes[del_key]

        return del_key

    def mutate_delete_connection(self):
        if self.connections:
            key = choice(list(self.connections.keys()))
            del self.connections[key]

    def distance(self, other, config):
        """
        Returns the genetic distance between this genome and the other. This distance value
        is used to compute genome compatibility for speciation.
        """

        # Compute node gene distance component.
        node_distance = 0.0
        if self.nodes or other.nodes:
            disjoint_nodes = 0
            for k2 in other.nodes:
                if k2 not in self.nodes:
                    disjoint_nodes += 1

            for k1, n1 in self.nodes.items():
                n2 = other.nodes.get(k1)
                if n2 is None:
                    disjoint_nodes += 1
                else:
                    # Homologous genes compute their own distance value.
                    node_distance += n1.distance(n2, config)

            max_nodes = max(len(self.nodes), len(other.nodes))
            node_distance = (node_distance +
                             (config.compatibility_disjoint_coefficient *
                              disjoint_nodes)) / max_nodes

        # Compute connection gene differences.
        connection_distance = 0.0
        if self.connections or other.connections:
            disjoint_connections = 0
            for k2 in other.connections:
                if k2 not in self.connections:
                    disjoint_connections += 1

            for k1, c1 in self.connections.items():
                c2 = other.connections.get(k1)
                if c2 is None:
                    disjoint_connections += 1
                else:
                    # Homologous genes compute their own distance value.
                    connection_distance += c1.distance(c2, config)

            max_conn = max(len(self.connections), len(other.connections))
            connection_distance = (connection_distance +
                                   (config.compatibility_disjoint_coefficient *
                                    disjoint_connections)) / max_conn

        distance = node_distance + connection_distance
        return distance

    def size(self):
        """
        Returns genome 'complexity', taken to be
        (number of nodes, number of enabled connections)
        """
        num_enabled_connections = sum([1 for cg in self.connections.values() if cg.enabled])
        return len(self.nodes), num_enabled_connections

    def __str__(self):
        s = "Key: {0}\nFitness: {1}\nNodes:".format(self.key, self.fitness)
        for k, ng in self.nodes.items():
            s += "\n\t{0} {1!s}".format(k, ng)
        s += "\nConnections:"
        connections = list(self.connections.values())
        connections.sort()
        for c in connections:
            s += "\n\t" + str(c)
        return s

    @staticmethod
    def create_node(config, node_id):
        node = config.node_gene_type(node_id)
        node.init_attributes(config)
        return node

    @staticmethod
    def create_connection(config, input_id, output_id):
        connection = config.connection_gene_type((input_id, output_id))
        connection.init_attributes(config)
        return connection

    def connect_fs_neat_nohidden(self, config):
        """
        Randomly connect one input to all output nodes
        (FS-NEAT without connections to hidden, if any).
        Originally connect_fs_neat.
        """
        input_id = choice(config.input_keys)
        for output_id in config.output_keys:
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection

    def connect_fs_neat_hidden(self, config):
        """
        Randomly connect one input to all hidden and output nodes
        (FS-NEAT with connections to hidden, if any).
        """
        input_id = choice(config.input_keys)
        others = [i for i in self.nodes if i not in config.input_keys]
        for output_id in others:
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection

    def compute_full_connections(self, config, direct):
        """
        Compute connections for a fully-connected feed-forward genome--each
        input connected to all hidden nodes
        (and output nodes if ``direct`` is set or there are no hidden nodes),
        each hidden node connected to all output nodes.
        (Recurrent genomes will also include node self-connections.)
        """
        hidden = [i for i in self.nodes if i not in config.output_keys]
        output = [i for i in self.nodes if i in config.output_keys]
        connections = []
        if hidden:
            for input_id in config.input_keys:
                for h in hidden:
                    connections.append((input_id, h))
            for h in hidden:
                for output_id in output:
                    connections.append((h, output_id))
        if direct or (not hidden):
            for input_id in config.input_keys:
                for output_id in output:
                    connections.append((input_id, output_id))

        # For recurrent genomes, include node self-connections.
        if not config.feed_forward:
            for i in self.nodes:
                connections.append((i, i))

        return connections

    def connect_full_nodirect(self, config):
        """
        Create a fully-connected genome
        (except without direct input-output unless no hidden nodes).
        """
        for input_id, output_id in self.compute_full_connections(config, False):
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection

    def connect_full_direct(self, config):
        """ Create a fully-connected genome, including direct input-output connections. """
        for input_id, output_id in self.compute_full_connections(config, True):
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection

    def connect_partial_nodirect(self, config):
        """
        Create a partially-connected genome,
        with (unless no hidden nodes) no direct input-output connections."""
        assert 0 <= config.connection_fraction <= 1
        all_connections = self.compute_full_connections(config, False)
        shuffle(all_connections)
        num_to_add = int(round(len(all_connections) * config.connection_fraction))
        for input_id, output_id in all_connections[:num_to_add]:
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection

    def connect_partial_direct(self, config):
        """
        Create a partially-connected genome,
        including (possibly) direct input-output connections.
        """
        assert 0 <= config.connection_fraction <= 1
        all_connections = self.compute_full_connections(config, True)
        shuffle(all_connections)
        num_to_add = int(round(len(all_connections) * config.connection_fraction))
        for input_id, output_id in all_connections[:num_to_add]:
            connection = self.create_connection(config, input_id, output_id)
            self.connections[connection.key] = connection



from ..graphs import required_for_output, feed_forward_layers

class RecurrentNetwork(object):
    def __init__(self, inputs, outputs, node_evals):
        self.input_nodes = inputs
        self.output_nodes = outputs
        self.node_evals = node_evals

        self.values = [{}, {}]
        for v in self.values:
            for k in list(inputs) + list(outputs):
                v[k] = 0.0

            for node, ignored_activation, ignored_aggregation, ignored_bias, ignored_response, links in self.node_evals:
                v[node] = 0.0
                for i, w in links:
                    v[i] = 0.0
        self.active = 0

    def reset(self):
        self.values = [dict((k, 0.0) for k in v) for v in self.values]
        self.active = 0

    def activate(self, inputs):
        if len(self.input_nodes) != len(inputs):
            raise RuntimeError("Expected {0:n} inputs, got {1:n}".format(len(self.input_nodes), len(inputs)))

        ivalues = self.values[self.active]
        ovalues = self.values[1 - self.active]
        self.active = 1 - self.active

        for i, v in zip(self.input_nodes, inputs):
            ivalues[i] = v
            ovalues[i] = v

        for node, activation, aggregation, bias, response, links in self.node_evals:
            node_inputs = [ivalues[i] * w for i, w in links]
            s = aggregation(node_inputs)
            ovalues[node] = activation(bias + response * s)

        return [ovalues[i] for i in self.output_nodes]

    @staticmethod
    def create(genome, config):
        """ Receives a genome and returns its phenotype (a RecurrentNetwork). """
        genome_config = config.genome_config
        required = required_for_output(genome_config.input_keys, genome_config.output_keys, genome.connections)

        # Gather inputs and expressed connections.
        node_inputs = {}
        for cg in genome.connections.values():
            if not cg.enabled:
                continue

            i, o = cg.key
            if o not in required and i not in required:
                continue

            if o not in node_inputs:
                node_inputs[o] = [(i, cg.weight)]
            else:
                node_inputs[o].append((i, cg.weight))

        node_evals = []
        for node_key, inputs in node_inputs.items():
            node = genome.nodes[node_key]
            activation_function = genome_config.activation_defs.get(node.activation)
            aggregation_function = genome_config.aggregation_function_defs.get(node.aggregation)
            node_evals.append((node_key, activation_function, aggregation_function, node.bias, node.response, inputs))

        return RecurrentNetwork(genome_config.input_keys, genome_config.output_keys, node_evals)


class MLRecurrentNetwork(object):
    def __init__(self, inputs, outputs, node_evals):
        self.input_nodes = inputs
        self.output_nodes = outputs
        self.node_evals = node_evals

        self.values = [{}, {}]
        r_idx0 = -len(inputs) - 1
        for v in self.values:
            for k in inputs + outputs:
                v[k] = 0.0

            for node, ignored_activation, ignored_aggregation, ignored_bias, ignored_response, links in self.node_evals:
                v[node] = 0.0
                for i, w in links:
                    if i > 0:  # Not a recurrent edge
                        v[i] = 0.0
                    elif i not in inputs:
                        v[-i+r_idx0] = 0
        self.active = 0

    def reset(self):
        self.values = [dict((k, 0.0) for k in v) for v in self.values]
        self.active = 0

    def activate(self, inputs):
        if len(self.input_nodes) != len(inputs):
            raise RuntimeError("Expected {0:n} inputs, got {1:n}".format(len(self.input_nodes), len(inputs)))

        t_m1_val = self.values[self.active]
        t_val = self.values[1 - self.active]
        self.active = 1 - self.active

        for i, v in zip(self.input_nodes, inputs):
            t_m1_val[i] = v
            t_val[i] = v

        r_idx0 = -len(inputs) - 1
        for node, activation, aggregation, bias, response, links in self.node_evals:
            node_inputs = [t_m1_val[-i+r_idx0] * w if i <= r_idx0 else t_val[i] for i, w in links]
            s = aggregation(node_inputs)
            t_val[node] = activation(bias + response * s)

        return [t_val[i] for i in self.output_nodes]

    @staticmethod
    def create(genome, config):
        """ Receives a genome and returns its phenotype (a MLRecurrentNetwork). """
        genome_config = config.genome_config
        connections = [cg.key for cg in genome.connections.values() if cg.enabled]

        layers = \
            feed_forward_layers(config.genome_config.input_keys, config.genome_config.output_keys, connections, True)

        node_evals = []
        for layer in layers:
            for node in layer:
                inputs = []
                for conn_key in connections:
                    inode, onode = conn_key
                    if onode == node:
                        cg = genome.connections[conn_key]
                        inputs.append((inode, cg.weight))

                ng = genome.nodes[node]
                aggregation_function = config.genome_config.aggregation_function_defs.get(ng.aggregation)
                activation_function = config.genome_config.activation_defs.get(ng.activation)
                node_evals.append((node, activation_function, aggregation_function, ng.bias, ng.response, inputs))

        return MLRecurrentNetwork(genome_config.input_keys, genome_config.output_keys, node_evals)



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

from collections import defaultdict
import importlib
import math
import re

from .datasets import get_dataset
from .utils import (
    arange,
    print_once,
    cast_ray,
    display,
    Line,
    Point,
    degrees_to_world,
    distance,
    intersect,
    rotate_around,
    PI_OVER_180,
    PI_OVER_2,
    ONE80_OVER_PI,
    TWO_PI,
    world_to_degrees,
)

from aitk.utils import Color

class Robot:
    """
    The Robot class is the base class for all robots.

    Typically, you would use an existing robot subclass:

    ```python
    from aitk.robots import Scribbler, Vehicle

    robot1 = Scribbler()
    robot2 = Vehicle()
    ```

    After creating a robot, you can then add devices to it,
    and then add it to a world.

    ```python
    from aitk.robots import World, Camera

    world = World(200, 200)
    robot1.add_device(Camera())

    world.add_robot(robot1)
    world.add_robot(robot2)
    ```
    """
    def __init__(
        self,
        x=0,
        y=0,
        a=0,
        color="red",
        name="Robbie",
        do_trace=True,
        height=0.25,
        max_trace_length=10,
        **kwargs
    ):
        """
        A simulated robot.

        Option | Type | Description | Default
        ------ | ---- | ----------- | -------
        x      | int  | starting location in the horizontal direction. Leave 0 to place in a random location. | 0
        y      | int  | starting location in the horizontal direction. Leave 0 to place in a random location. | 0
        a      | float | starting angle in degrees. | 0
        color  | str  | the name of a color for the robot | "red"
        name   | str  | a name to give your robot | "Robbie"
        do_trace | bool | should the robot leave a trace? | True
        height | int  | height of robot use number < 1 | 0.25
        max_trace_length | int | max length of trace, in seconds | 10
        state  | dict | serializable memory for a robot | None
        va     | float | angle velocity | None
        vx     | float | x velocity | None
        vy     | float | y velocity | None
        tva    | float | angle target velocity | None
        tvx    | float | x target velocity | None
        tvy    | float | y target velocity | None
        va_max | float | angle max velocity | None
        vx_max | float | x max velocity | None
        vy_max | float | y max velocity | None
        va_ramp | float | angle linear acceleration | None
        vx_ramp | float | x linear acceleration | None
        vy_ramp | float | y linear acceleration | None
        image_data | dataset-name, index | to use a 3D set of images | None
        body   | List | data structure that defines a robot body | None
        devices | List | list of serialized devices | None
        """
        # Get the args:
        config = {
            "x": x,
            "y": y,
            "a": a, # degrees in the config file
            "color": color,
            "name": name,
            "do_trace": do_trace,
            "height": height,
            "max_trace_length": max_trace_length,
        }
        # Update from the kwargs:
        config.update(kwargs)
        self.world = None
        self._devices = []
        self._initialize()
        self.from_json(config)

    def get_distances(self, *positions, scaled=False):
        if len(positions) == 0:
            positions = []
            for p in range(len(self)):
                if self[p].__class__.__name__ == "RangeSensor":
                    positions.append(p)
        if scaled:
            return [self[p].get_distance() / self[p].get_max() for p in positions]
        else:
            return [self[p].get_distance() for p in positions]

    def get_readings(self, *positions):
        return [self[p].get_reading() for p in positions]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._devices[item]
        elif isinstance(item, str):
            type_map = {}  # mapping of base to count
            search_groups = re.match(r"(.*)-(\d*)", item)
            if search_groups:
                search_type = search_groups[1].lower()
                if search_groups[2].isdigit():
                    search_index = int(search_groups[2])
                else:
                    search_type = item
                    search_index = 1
            else:
                search_type = item.lower()
                search_index = 1
            for device in self._devices:
                # update type_map
                device_type = device.type.lower()
                device_name = device.name.lower()
                device_index = None
                if "-" in device_type:
                    device_prefix, device_index = device_type.rsplit("-", 1)
                    if device_index.isdigit():
                        device_type = device_prefix
                        device_index = int(device_index)
                    else:
                        device_index = 1
                if device_type not in type_map:
                    type_map[device_type] = 1
                else:
                    type_map[device_type] += 1
                if device_index is None:
                    device_index = type_map[device_type]
                if search_type == device_type and search_index == device_index:
                    return device
                if search_type == device_name:
                    return device
        return None

    def __len__(self):
        return len(self._devices)

    def __repr__(self):
        if self.world is None:
            return "<Robot(name=%r, unconnected)>" % (self.name,)
        else:
            return "<Robot(name=%r, x=%s, y=%s, a=%s)>" % (
                self.name,
                round(self.x, 2),
                round(self.y, 2),
                round(world_to_degrees(self.a), 2),
            )

    def _initialize(self):
        """
        Initialize the robot properties.
        """
        self.name = "Robbie"
        self.state = defaultdict(int)
        self._set_color("red")
        self.eat_food_distance = 20
        self.do_trace = True
        self.trace = []
        self.text_trace = []
        self.pen_trace = []
        self.pen = (None, 0)
        self.body = []
        self.max_trace_length = 10  # seconds
        self.x = 0  # cm
        self.y = 0  # cm
        self.height = 0.25
        self.a = degrees_to_world(0)  # radians
        self.vx = 0.0  # velocity in x direction, CM per second
        self.vy = 0.0  # velocity in y direction, degrees per second
        self.va = 0.0  # turn velocity
        self.tvx = 0.0
        self.tvy = 0.0
        self.tva = 0.0
        self.va_ramp = 1.0  # seconds to reach max speed
        self.vx_ramp = 1.0  # seconds to reach max speed
        self.vy_ramp = 1.0  # seconds to reach max speed
        self.vx_max = 2.0  # CM/SEC
        self.va_max = math.pi * 0.90  # RADIANS/SEC
        self.vy_max = 2.0  # CM/SEC
        self.stalled = False
        # Slight box around center, in case no
        # body:
        self._bounding_lines = [
            Line(Point(-1, -1), Point(1, -1)),
            Line(Point(1, -1), Point(1, 1)),
            Line(Point(1, 1), Point(-1, 1)),
            Line(Point(-1, 1), Point(-1, -1)),
        ]
        self.state = defaultdict(int)
        self.image_data = []
        self.get_dataset_image = None
        self._boundingbox = []
        self.radius = 0.0
        self.food_eaten = 0
        self._watcher = None
        self._init_boundingbox()

    def from_json(self, config):
        """
        Load a robot from a JSON config dict.

        Args:
            config (dict): a config dict
        """
        DEVICES = importlib.import_module("aitk.robots.devices")
        valid_keys = set([
            "name", "state", "do_trace", "va", "vx", "vy",
            "tva", "tvx", "tvy", "x", "y", "a", "va_max",
            "vx_max", "vy_max", "va_ramp", "vx_ramp", "vy_ramp",
            "image_data", "height", "color", "max_trace_length",
            "body", "devices",
        ])
        config_keys = set(list(config.keys()))
        extra_keys = config_keys - valid_keys

        if len(extra_keys) > 0:
            raise TypeError("invalid key(s) for robot config: %r" % extra_keys)

        if "name" in config:
            self.name = config["name"]

        if "state" in config:
            self.state = defaultdict(int)
            self.state.update(config["state"])

        if "do_trace" in config:
            self.do_trace = config["do_trace"]

        if "va" in config:
            self.va = config["va"]
        if "vx" in config:
            self.vx = config["vx"]
        if "vy" in config:
            self.vy = config["vy"]

        if "tva" in config:
            self.tva = config["tva"]
        if "tvx" in config:
            self.tvx = config["tvx"]
        if "tvy" in config:
            self.tvy = config["tvy"]

        if "va_max" in config:
            self.va_max = config["va_max"]
        if "vx_max" in config:
            self.vx_max = config["vx_max"]
        if "vy_max" in config:
            self.vy_max = config["vy_max"]

        if "va_ramp" in config:
            self.va_ramp = config["va_ramp"]
        if "vx_ramp" in config:
            self.vx_ramp = config["vx_ramp"]
        if "vy_ramp" in config:
            self.vy_ramp = config["vy_ramp"]

        if "x" in config:
            self.x = config["x"]
        if "y" in config:
            self.y = config["y"]
        if "a" in config:
            self.a = degrees_to_world(config["a"])

        if "image_data" in config:
            self.image_data = config["image_data"]  # ["dataset", index]
        if len(self.image_data) == 0:
            self.get_dataset_image = None
        else:
            self.get_dataset_image = get_dataset(self.image_data[0])

        if "height" in config:
            self.height = config["height"]  # ratio, 0 to 1 of height

        if "color" in config:
            self._set_color(config["color"])

        if "max_trace_length" in config:
            self.max_trace_length = config["max_trace_length"]

        if "body" in config:
            self.body[:] = config["body"]
            self._init_boundingbox()

        if "devices" in config:
            # FIXME: raise if lengths/types don't match
            for i, deviceConfig in enumerate(config["devices"]):
                if i < len(self):
                    if self[i].__class__.__name__ == deviceConfig["class"]:
                        # already a device, let's reuse it:
                        device = self[i]
                        device.initialize()
                        device.from_json(deviceConfig)
                    else:
                        raise Exception(
                            "can't use reset; config changed; use load_world"
                        )
                else:
                    device = None
                    try:
                        device_class = getattr(DEVICES, deviceConfig["class"])
                        device = device_class(**deviceConfig)
                    except Exception:
                        raise Exception(
                            "Failed to create device: %s(**%s)"
                            % (deviceConfig["class"], deviceConfig)
                        )
                    if device:
                        self.add_device(device)

    def summary(self):
        """
        Prints a summary of information about the robot.
        """
        if len(self._devices) == 0:
            print("  This robot has no devices.")
        else:
            for i, device in enumerate(self._devices):
                print(
                    "      robot[%s or %r or %r]: %r"
                    % (i, device.type, device.name, device)
                )
            print("  " + ("-" * 25))

    def get_widget(self, size=None, show_robot=None, attributes=None):
        """
        Get the robot widget.

        Args:
            size (int): size in pixels around robot
            show_robot (bool): show picture of robot
            attributes (list): items to include, or "all"
        """
        from .watchers import RobotWatcher

        if self._watcher is None:
            size = size if size is not None else 100
            show_robot = show_robot if show_robot is not None else True
            attributes = attributes if attributes is not None else "all"
            self._watcher = RobotWatcher(self, size=size,
                                         show_robot=show_robot,
                                         attributes=attributes)
            self.world._watchers.append(self._watcher)
        else:
            self._watcher.set_arguments(size=size,
                                        show_robot=show_robot,
                                        attributes=attributes)
            self._watcher.draw()

        return self._watcher.get_widget()

    def watch(self, size=None, show_robot=None, attributes=None):
        """
        Watch the robot stats with live updates.

        Args:
            size (int): size in pixels around robot
            show_robot (bool): show picture of robot
            attributes (List): items to include, or "all"
        """
        widget = self.get_widget(
            size=size,
            show_robot=show_robot,
            attributes=attributes,
        )
        display(widget)

    def get_image(self, size=100):
        """
        Get an image of the robot.

        Args:
            size (int): size in pixels around robot
        """
        picture = self.world.get_image()
        start_x = round(
            max(self.x * self.world.scale - size / 2, 0)
        )
        start_y = round(
            max(self.y * self.world.scale - size / 2, 0)
        )
        rectangle = (
            start_x,
            start_y,
            min(
                start_x + size, self.world.width * self.world.scale
            ),
            min(
                start_y + size,
                self.world.height * self.world.scale,
            ),
        )
        picture = picture.crop(rectangle)
        return picture

    def display(self, size=100):
        """
        Display the robot's image.

        Args:
            size (int): the size of the width and height of the
                image
        """
        image = self.get_image(size=size)
        display(image)

    def set_max_trace_length(self, seconds):
        """
        Set the max length of trace, in seconds.

        Args:
            seconds (float): the length of trace in seconds
        """
        self.max_trace_length = seconds

    def set_color(self, color):
        """
        Set the color of a robot, and its trace.

        Args:
            color (str): the color of the robot
        """
        self._set_color(color)

    def _set_color(self, color):
        if not isinstance(color, Color):
            self.color = Color(color)
        else:
            self.color = color
        self.trace_color = Color(
            self.color.red * 0.75, self.color.green * 0.75, self.color.blue * 0.75,
        )

    def set_pose(self, x=None, y=None, a=None, clear_trace=True,
                 show=True):
        """
        Set the pose of the robot. a is in degrees.

        Args:
            x (int): the x coordinate of the robot
            y (int): the y coordinate of the robot
            a (int): the angle of the robot in degrees (zero is to the left, 90 is up)
            clear_trace (bool): if True, clear the robot's trace
            show (bool): if True, show the updated world

        Note: the robot must be in a world.
        """
        if self.world is None:
            raise Exception(
                "This robot is not in a world; add to world before setting pose"
            )
        else:
            if a is not None:
                a = degrees_to_world(a)
            self._set_pose(x, y, a, clear_trace)
            # Set all velocities to zero
            self.vx = 0.0  # velocity in x direction, CM per second
            self.vy = 0.0  # velocity in y direction, degrees per second
            self.va = 0.0  # turn velocity
            self.tvx = 0.0
            self.tvy = 0.0
            self.tva = 0.0
            self.stalled = False
            # Save the robot's pose to the config
            self.world.update(show=show)
            self.world.save()

    def set_random_pose(self, clear_trace=True, show=True):
        """
        Set the pose of the robot to a random open location in the world.

        Args:
            clear_trace (bool): if True, clear the trace
            show (bool): if True, show the updated world

        Note: the robot must be in a world.
        """
        if self.world is None:
            raise Exception(
                "This robot is not in a world; add robot to world before calling set_random_pose"
            )
        else:
            # Direction is in radians, in world coordinates:
            x, y, a = self.world._find_random_pose(self)
            self._set_pose(x, y, a, clear_trace)
            # Set all velocities to zero
            self.vx = 0.0  # velocity in x direction, CM per second
            self.vy = 0.0  # velocity in y direction, degrees per second
            self.va = 0.0  # turn velocity
            self.tvx = 0.0
            self.tvy = 0.0
            self.tva = 0.0
            self.stalled = False
            # Save the robot's pose to the config
            self.world.update(show=show)
            self.world.save()

    def _set_pose(self, x=None, y=None, a=None, clear_trace=True):
        """
        Set the pose of the robot. direction (a) is in radians.

        a: direction is in radians, in world coordinates.
        """
        if clear_trace:
            self.trace[:] = []
            self.text_trace[:] = []
            self.pen_trace[:] = []
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if a is not None:
            self.a = a

    def del_device(self, device):
        """
        Remove a device from a robot.

        Args:
            device (Device): the device to remove
        """
        if isinstance(device, (str, int)):
            device = self[device]
        if device in self._devices:
            device.robot = None
            self._devices.remove(device)
        else:
            raise Exception("Device %r is not on robot." % device)

    def add_device(self, device):
        """
        Add a device to a robot.

        Args:
            device (Device): the device to add
        """
        if device not in self._devices:
            device.robot = self
            self._devices.append(device)
            if self.world is not None:
                self.world.update()  # request draw
        else:
            raise Exception("Can't add the same device to a robot more than once.")

    def add_device_ring(self, device_class, distance_from_center,
                        start_degree, stop_degree, count, **kwargs):
        """
        Adds a ring of devices at a given distance from the center of
        the robot.

        Args:
            device_class (class): a class or function that receives position,
                a, and kwargs, and returns a device
            distance_from_center (float): in CM
            start_degree (float): angle of first device (0 points right)
            stop_degree (float): angle of stop degree (counter clockwise)
            count (int): number of sensors to add

        Note: use kwargs additional args to pass to device_class

        Example:

        ```python
        >>> robot.add_device_ring(RangeSensor, 10, 0, 360, 6, width=20)
        ```
        """
        if isinstance(device_class, str):
            DEVICES = importlib.import_module("aitk.robots.devices")
            device_class = getattr(DEVICES, device_class)

        if stop_degree < start_degree:
            stop_degree += 360

        span = stop_degree - start_degree
        if span == 360:
            step_angle = span / count
        else:
            step_angle = span / (count - 1)

        angle = start_degree
        for i in range(count):
            x, y = rotate_around(0, 0, distance_from_center, -angle * PI_OVER_180)
            self.add_device(device_class(position=(x, y), a=angle, **kwargs))
            angle += step_angle

    def to_json(self):
        """
        Get this robot as a JSON config file.
        """
        robot_json = {
            "name": self.name,
            "state": self.state,
            "va": self.va,
            "vx": self.vx,
            "vy": self.vy,
            "tva": self.tva,
            "tvx": self.tvx,
            "tvy": self.tvy,
            "va_max": self.va_max,
            "vx_max": self.vx_max,
            "vy_max": self.vy_max,
            "va_ramp": self.va_ramp,
            "vx_ramp": self.vx_ramp,
            "vy_ramp": self.vy_ramp,
            "x": self.x,
            "y": self.y,
            "a": world_to_degrees(self.a),
            "image_data": self.image_data,
            "height": self.height,
            "color": str(self.color),
            "max_trace_length": self.max_trace_length,
            "body": self.body,
            "devices": [device.to_json() for device in self._devices],
            "do_trace": self.do_trace,
        }
        return robot_json

    def motors(self, left, right):
        """
        A move function that takes desired motor values
        and converts to trans and rotate.

        Args:
            left (float): the speed to move the left wheel
            right (float): the speed to move the right wheel
        """
        trans = (right + left) / 2.0
        rotate = (right - left) / 2.0
        self.move(trans, rotate)

    def imotors(self, left, right):
        """
        A move function that takes desired motor values
        and converts to trans and rotate. Uses instantaneous
        move.

        Args:
            left (float): the speed to move the left wheel
            right (float): the speed to move the right wheel
        """
        trans = (right + left) / 2.0
        rotate = (right - left) / 2.0
        self.imove(trans, rotate)

    def move(self, translate, rotate):
        """
        Set the target translate and rotate velocities.

        Args:
            translate (float): the speed to move the robot forward/backward
            rotate (float): the speed to rotate the robot left/right

        Args should be between -1 and 1.
        """
        # values between -1 and 1
        # compute target velocities
        if self.world is not None:
            if self.world.status != "running":
                print_once("This world is not running")
        if translate is not None:
            self.tvx = round(translate * self.vx_max, 1)
        if rotate is not None:
            self.tva = round(rotate * self.va_max, 1)

    def imove(self, translate, rotate):
        """
        Instantaneously set target speeds (zero momentum).

        Args:
            translate (float): the speed to move the robot forward/backward
            rotate (float): the speed to rotate the robot left/right

        Args should be between -1 and 1.
        """
        # values between -1 and 1
        # compute target velocities
        if self.world is not None:
            if self.world.status != "running":
                print_once("This world is not running")
        if translate is not None:
            self.tvx = self.vx = round(translate, 1)
        if rotate is not None:
            self.tva = self.va = round(rotate, 1)

    def translate(self, translate):
        """
        Set the target translate velocity.

        Args:
            translate (float): the speed to move the robot forward/backward

        Arg should be between -1 and 1.
        """
        # values between -1 and 1
        self.tvx = round(translate * self.vx_max, 1)

    def forward(self, translate):
        """
        Set the target translate velocity.

        Args:
            translate (float): the speed to move the robot forward

        Arg should be between 0 and 1, inclusive.
        """
        if 0 <= translate <= 1:
            self.tvx = round(translate * self.vx_max, 1)
        else:
            print("forward value is out of range; should be between 0 and 1, inclusive")

    def backward(self, translate):
        """
        Set the target translate velocity.

        Args:
            translate (float): the speed to move the robot backward

        Translate should be between 0 and 1, inclusive.
        """
        if 0 <= translate <= 1:
            self.tvx = round(-translate * self.vx_max, 1)
        else:
            print("backward value is out of range; should be between 0 and 1, inclusive")

    def reverse(self):
        """
        Flip the target x and a velocities from negative to
        positive or from positive to negative.
        """
        self.tvx = -self.tvx
        self.tva = -self.tva

    def rotate(self, rotate):
        """
        Set the target rotate velocity.

        Args:
            rotate (float): the speed to rotate the robot

        Arg should be between -1 and 1.
        """
        # values between -1 and 1
        self.tva = rotate * self.va_max

    def stop(self):
        """
        Set the target velocities to zeros.
        """
        self.tvx = 0.0
        self.tvy = 0.0
        self.tva = 0.0

    def has_image(self):
        """
        Does this robot have an associated 3D set of images from
        a dataset?
        """
        return self.get_dataset_image is not None

    def eat(self):
        """
        If the robot is close enough to food, then this
        will eat it, removing it from the world, and requesting
        a redraw. Returns True if successfully eaten, and False
        otherwise.

        Note: it must be within robot.eat_food_distance
        """
        success = False
        if self.world is not None:
            for food in [f for f in self.world._food if f.state == "on"]:
                if distance(self.x, self.y, food.x, food.y) <= self.eat_food_distance:
                    success = True
                    self.world._event("food-off", robot=self, food=food)
        return success

    def speak(self, text=None):
        """
        Show some text in the robot's speech bubble.

        Args:
            text (str): the text to show; use None to clear

        Note: not for use in a robot in a recorder.
        """
        if self.world:
            if self.world._recording:
                if len(self.text_trace) > 0:
                    # If same as last, don't add again
                    if self.text_trace[-1][1] != text:
                        self.text_trace.append((self.world.time, text))
                else:
                    self.text_trace.append((self.world.time, text))
            else:
                self.text_trace[:] = [(self.world.time, text)]

    def pen_down(self, color=None, radius=1):
        """
        Put the pen down to change the color of the background image.

        Args:
            color (str): the pen color to draw
            radius (int): the size of the dot to draw

        Note: not for use in a robot in a recorder.
        """
        from PIL import Image

        color = color if color is not None else self.color
        self.pen = (Color(color), radius)
        if self.world is not None:
            if self.world._ground_image is None:
                image = Image.new("RGBA", (self.world.width, self.world.height),
                                  color="white")
                filename = "ground_image.png"
                image.save(filename)
                self.world.set_ground_image(filename)

    def pen_up(self):
        """
        Put the pen up to stop changing the color of the background image.

        Note: not for use in a robot in a recorder.
        """
        self.pen = (None, 0)

    def get_max_trace_length(self):
        """
        Get the max length of the trace in seconds.
        """
        return self.max_trace_length

    def get_image_3d(self, degrees):
        """
        Return the 3D image at the proper angle.

        Args:
            degrees (int): the angle of the image to get
        """
        return self.get_dataset_image(self.image_data[1], degrees)

    def get_time(self):
        """
        Get the clock time of the world.
        """
        if self.world:
            return self.world.time

    def get_pose(self):
        """
        Get the pose of the robot (x, y, a) where a (direction)
        is in degrees.
        """
        return (self.x, self.y, world_to_degrees(self.a))

    def get_velocity(self, target=False):
        """
        Get the current (or target) translate and rotate velocities
        of the robot.

        Args:
            target (bool): get the target velocities if True; get the actual
                velocities otherwise
        """
        if not target:
            return (self.vx / self.vx_max, self.va / self.va_max)
        else:
            return (self.tvx / self.vx_max, self.tva / self.va_max)

    def cast_ray(self, x1, y1, a, maxRange, x2=None, y2=None, ignore_robots=None):
        """
        Cast a ray into this world and see what it hits.

        Args:
            x1 (int): the x coordinate of ray origin
            y1 (int): the y coordinate of ray origin
            a (int): the angle of the ray to cast
            maxRange (float): the length of ray to cast
            x2 (int): the x coordinate of the endpoint
            y2 (int): the y coordinate of the endpoint
            ignore_robots (List): list of robots to ignore

        Returns list of hits, furthest away first (back to front)
        """
        return cast_ray(self.world, self, x1, y1, a, maxRange, x2, y2, ignore_robots)

    def _init_boundingbox(self):
        # First, find min/max points around robot (assumes box):
        min_x = float("inf")
        max_x = float("-inf")
        min_y = float("inf")
        max_y = float("-inf")
        max_dist = float("-inf")

        if len(self.body) > 0 and self.body[0][0] == "polygon":
            # "polygon", color, points
            for point in self.body[0][2]:
                min_x = min(min_x, point[0])
                min_y = min(min_y, point[1])
                max_x = max(max_x, point[0])
                max_y = max(max_y, point[1])
                max_dist = max(max_dist, distance(0, 0, point[0], point[1]))

        if (
            min_x == float("inf")
            or min_y == float("inf")
            or max_x == float("-inf")
            or max_y == float("-inf")
        ):
            return

        self._boundingbox = [min_x, min_y, max_x, max_y]
        self.radius = max_dist
        ps = self._compute_boundingbox(self.x, self.y, self.a)
        self._update_boundingbox(*ps)

    def _compute_boundingbox(self, px, py, pa):
        # Compute position in real world with respect to x, y, a:
        if len(self._boundingbox) == 4:
            min_x, min_y, max_x, max_y = self._boundingbox
            ps = []
            for x, y in [
                (min_x, max_y),  # 4
                (min_x, min_y),  # 1
                (max_x, min_y),  # 2
                (max_x, max_y),  # 3
            ]:
                dist = distance(0, 0, x, y)
                angle = math.atan2(-x, y)
                p = rotate_around(px, py, dist, pa + angle + PI_OVER_2)
                ps.append(p)
            return ps
        else:
            raise ValueError("robot %r does not have a body" % self.name)

    def reset(self):
        """
        Reset the robot's internal stuff. Typically, called from the world.
        """
        self.trace[:] = []
        self.text_trace[:] = []
        self.pen_trace[:] = []

    def _restore_boundingbox(self):
        self._update_boundingbox(*self._last_boundingbox)

    def _update_boundingbox(self, p1, p2, p3, p4):
        self._last_boundingbox = [
            self._bounding_lines[0].p1.copy(),  # p1
            self._bounding_lines[0].p2.copy(),  # p2
            self._bounding_lines[1].p2.copy(),  # p3
            self._bounding_lines[2].p2.copy(),  # p4
        ]
        self._bounding_lines[0].p1.x = p1[0]
        self._bounding_lines[0].p1.y = p1[1]
        self._bounding_lines[0].p2.x = p2[0]
        self._bounding_lines[0].p2.y = p2[1]

        self._bounding_lines[1].p1.x = p2[0]
        self._bounding_lines[1].p1.y = p2[1]
        self._bounding_lines[1].p2.x = p3[0]
        self._bounding_lines[1].p2.y = p3[1]

        self._bounding_lines[2].p1.x = p3[0]
        self._bounding_lines[2].p1.y = p3[1]
        self._bounding_lines[2].p2.x = p4[0]
        self._bounding_lines[2].p2.y = p4[1]

        self._bounding_lines[3].p1.x = p4[0]
        self._bounding_lines[3].p1.y = p4[1]
        self._bounding_lines[3].p2.x = p1[0]
        self._bounding_lines[3].p2.y = p1[1]

    def _deltav(self, tv, v, maxv, ramp, time_step):
        # max change occurs in how long:
        seconds = ramp
        # how much can we change in one time step?
        spt = seconds / time_step
        dv = maxv / spt  # change in one time step
        return min(max(tv - v, -dv), dv)  # keep in limit

    def _step(self, time_step):
        """
        Have the robot make one step in time. Check to see if it hits
        any obstacles.
        """
        # proposed acceleration:
        va = self.va + self._deltav(
            self.tva, self.va, self.va_max, self.va_ramp, time_step
        )
        vx = self.vx + self._deltav(
            self.tvx, self.vx, self.vx_max, self.vx_ramp, time_step
        )
        vy = self.vy + self._deltav(
            self.tvy, self.vy, self.vy_max, self.vy_ramp, time_step
        )
        # graphics offset:
        offset = PI_OVER_2
        # proposed positions:
        pa = self.a - va * time_step
        tvx = (
            vx * math.sin(-pa + offset)
            + vy * math.cos(-pa + offset) * time_step
        )
        tvy = (
            vx * math.cos(-pa + offset)
            - vy * math.sin(-pa + offset) * time_step
        )
        px = self.x + tvx
        py = self.y + tvy

        # check to see if collision
        # bounding box:
        p1, p2, p3, p4 = self._compute_boundingbox(px, py, pa)
        # Set wall bounding boxes for collision detection:
        self._update_boundingbox(p1, p2, p3, p4)

        self.stalled = False
        # if intersection, can't move:
        for wall in self.world._walls:
            if wall.robot is self:  # if yourself, don't check for collision
                continue
            for line in wall.lines:
                w1 = line.p1
                w2 = line.p2
                if (
                    intersect(p1[0], p1[1], p2[0], p2[1], w1.x, w1.y, w2.x, w2.y)
                    or intersect(p2[0], p2[1], p3[0], p3[1], w1.x, w1.y, w2.x, w2.y)
                    or intersect(p3[0], p3[1], p4[0], p4[1], w1.x, w1.y, w2.x, w2.y)
                    or intersect(p4[0], p4[1], p1[0], p1[1], w1.x, w1.y, w2.x, w2.y)
                ):
                    self.stalled = True
                    break

        if not self.stalled:
            # if no intersection, make move
            self.va = va
            self.vx = vx
            self.vy = vy
            self.x = px
            self.y = py
            self.a = pa
        else:
            self._restore_boundingbox()
            # Adjust actual velocity
            self.va = 0
            self.vx = 0
            self.vy = 0

        # Devices:
        for device in self._devices:
            device._step(time_step)

        # Update history:
        if self.do_trace:
            self.trace.append((Point(self.x, self.y), self.a))

    def update(self, draw_list=None):
        """
        Update the robot, and devices.

        Args:
            draw_list (List): list of items to draw
        """
        # Wrapped worlds:
        wrapped = False
        if self.x < 0:
            self.x = self.world.width
            wrapped = True
        elif self.x > self.world.width:
            self.x = 0
            wrapped = True
        if self.y < 0:
            self.y = self.world.height
            wrapped = True
        elif self.y > self.world.height:
            self.y = 0
            wrapped = True

        if wrapped:
            self.trace.append(None)

        self._init_boundingbox()

        if self.world.debug and draw_list is not None:
            draw_list.append(("strokeStyle", (Color(255), 1)))
            draw_list.append(
                (
                    "draw_line",
                    (
                        self._bounding_lines[0].p1.x,
                        self._bounding_lines[0].p1.y,
                        self._bounding_lines[0].p2.x,
                        self._bounding_lines[0].p2.y,
                    ),
                )
            )

            draw_list.append(
                (
                    "draw_line",
                    (
                        self._bounding_lines[1].p1.x,
                        self._bounding_lines[1].p1.y,
                        self._bounding_lines[1].p2.x,
                        self._bounding_lines[1].p2.y,
                    ),
                )
            )

            draw_list.append(
                (
                    "draw_line",
                    (
                        self._bounding_lines[2].p1.x,
                        self._bounding_lines[2].p1.y,
                        self._bounding_lines[2].p2.x,
                        self._bounding_lines[2].p2.y,
                    ),
                )
            )

            draw_list.append(
                (
                    "draw_line",
                    (
                        self._bounding_lines[3].p1.x,
                        self._bounding_lines[3].p1.y,
                        self._bounding_lines[3].p2.x,
                        self._bounding_lines[3].p2.y,
                    ),
                )
            )

        # Devices:
        for device in self._devices:
            device.update(draw_list)

        # Update recording info:
        if self.world._recording:
            if len(self.pen_trace) > 0:
                # color of pen:
                if self.pen_trace[-1][1][0] != self.pen[0]:
                    self.pen_trace.append((self.world.time, self.pen))
                # same pen color, do nothing
            elif self.pen != (None, 0):
                self.pen_trace.append((self.world.time, self.pen))
            # else do nothing
        else:  # not recording
            if self.pen == (None, 0):
                self.pen_trace[:] = []
            else:
                self.pen_trace[:] = [(self.world.time, self.pen)]

        # Alter world:
        self._update_ground_image(self.world.time)
        return

    def _get_current_text(self, time):
        """
        Get the text for the specific time.
        """
        # FIXME: rewrite as binary search
        if len(self.text_trace) > 0:
            # find the last text that is after time
            if len(self.text_trace) == 1:
                return self.text_trace[0][1]
            for index in range(-1, -len(self.text_trace) - 1, -1):
                curr_time, curr_text = self.text_trace[index]
                if curr_time <= time:
                    return curr_text

    def _get_current_pen_color(self, time):
        # FIXME: rewrite as binary search
        if len(self.pen_trace) > 0:
            # find the last color that is after time
            if len(self.pen_trace) == 1:
                return self.pen_trace[0]
            for index in range(-1, -len(self.pen_trace) - 1, -1):
                data = self.pen_trace[index]
                # time, (color, radius)
                if data[0] <= time:
                    return data

    def _update_ground_image(self, world_time):
        data = self._get_current_pen_color(world_time)
        # time, (color, radius)
        if data is not None and data[1][0] is not None:
            self.world.set_ground_color_at(self.x, self.y, data[1])

    def draw(self, backend):
        """
        Draw the robot.

        Args:
            backend (Backend): the backend on which to draw
        """
        if self.do_trace:
            time_step = self.world.time_step if self.world is not None else 0.1
            if self.max_trace_length > 0:
                max_trace_length = int(1.0 / time_step * self.max_trace_length)
            else:
                max_trace_length = 0

            if max_trace_length == 0:
                data = []
            else:
                data = self.trace[-max_trace_length:]

            # None indicates a segment break
            if all(data): # no segments
                segments = [[(point[0], point[1]) for (point, direction) in data]]
            else:
                segments = []
                current = []
                for item in data:
                    if item is None:
                        segments.append(current)
                        current = []
                    else:
                        point, direction = item
                        current.append((point[0], point[1]))
                if current:
                    segments.append(current)

            for segment in segments:
                backend.draw_lines(
                    segment,
                    stroke_style=self.trace_color,
                )


            self.trace = data

        backend.pushMatrix()
        backend.translate(self.x, self.y)
        backend.rotate(self.a)

        # Draw first:
        for device in self._devices:
            if device.type == "bulb":
                device.draw(backend)
            elif device.type == "beacon":
                device.draw(backend)

        # body:
        for shape in self.body:
            shape_name, color, args = shape

            if self.stalled:
                backend.strokeStyle(Color(255), 1)
            else:
                backend.strokeStyle(Color(0), 1)

            if color is None:
                backend.set_fill(self.color)
            else:
                backend.set_fill(Color(color))

            if shape_name == "polygon":
                backend.draw_polygon(args)
            elif shape_name == "rectangle":
                backend.draw_rect(*args)
            elif shape_name == "ellipse":
                backend.draw_ellipse(*args)
            elif shape_name == "circle":
                backend.draw_circle(*args)
            elif shape_name == "line":
                backend.draw_line(*args)

            backend.noStroke()

        # Draw on top of robot:
        for device in self._devices:
            if device.type not in ["bulb", "beacon"]:
                device.draw(backend)

        backend.popMatrix()

        text = self._get_current_text(self.world.time)
        if text:
            backend.set_fill_style(Color(255))
            pad = 10
            box_pad = 5
            width = self.world._backend.char_width * len(text)
            height = 20
            if self.x - pad - width < 0:
                side = 1  # put on right
            else:
                side = -1  # put on left
            if self.y - height < 0:
                half = 1  # put on top
            else:
                half = -1  # put on bottom
            points = [
                (self.x, self.y),
                (self.x + pad * side, self.y + height / 2 * half),
                (self.x + pad * side, self.y + height * half),
                (self.x + (pad + width + pad) * side, self.y + height * half),
                (self.x + (pad + width + pad) * side, self.y),
                (self.x + pad * side, self.y),
                (self.x + pad * side, self.y + height / 4 * half),
            ]
            backend.set_stroke_style(Color(0))
            backend.set_fill_style(Color(255, 255, 255, 200))
            backend.draw_polygon(points)
            backend.set_fill_style(Color(0))
            if side == 1:  # right
                if half == 1:  # bottom
                    backend.text(text, self.x + (pad + box_pad), self.y + box_pad)
                else:  # top
                    backend.text(
                        text,
                        self.x + (pad + box_pad),
                        self.y - (pad + box_pad),
                    )
            else:  # left
                if half == 1:  # bottom
                    backend.text(text, self.x - pad - width - box_pad, self.y + box_pad)
                else:  # top
                    backend.text(
                        text,
                        self.x - pad - width - box_pad,
                        self.y - (pad + box_pad),
                    )


SCRIBBLER_CONFIG = {
    "body": [
        [
            "polygon",
            None,
            [
                [4.17, 5],
                [4.17, 6.67],
                [5.83, 5.83],
                [5.83, 5],
                [7.5, 5],
                [7.5, -5],
                [5.83, -5],
                [5.83, -5.83],
                [4.17, -6.67],
                [4.17, -5],
                [-4.17, -5],
                [-4.17, -6.67],
                [-5.83, -5.83],
                [-6.67, -5],
                [-7.5, -4.17],
                [-7.5, 4.17],
                [-6.67, 5],
                [-5.83, 5.83],
                [-4.17, 6.67],
                [-4.17, 5],
            ],
        ],
        ["rectangle", "black", [-3.33, -7.67, 6.33, 1.67]],
        ["rectangle", "black", [-3.33, 6, 6.33, 1.67]],
        ["polygon", "black", [[0, 0], [-2, 2], [2, 0], [-2, -2]]],
    ],
}

class Scribbler(Robot):
    def __init__(
        self,
        x=0,
        y=0,
        a=0,
        color="red",
        name="Scribbie",
        do_trace=True,
        height=0.25,
        max_trace_length=10,
        **kwargs
    ):
        config = {
            "x": x,
            "y": y,
            "a": a, # degrees in config file
            "color": color,
            "name": name,
            "do_trace": do_trace,
            "height": height,
            "max_trace_length": max_trace_length,
        }
        # First, get Scribbler defaults:
        defaults = SCRIBBLER_CONFIG.copy()
        # Then update from args:
        defaults.update(config)
        # Then update from kwargs:
        defaults.update(kwargs)
        super().__init__(**defaults)


VEHICLE_CONFIG = {
    "body": [
        [
            "polygon",
            None,
            [
                [5, 3.25],
                [5, -3.25],
                [-5, -3.25],
                [-5, 3.25],
            ],
        ],
        ["rectangle", "black", [-6, -3.25 - 1.5, 2.75, 1.5]],
        ["rectangle", "black", [-6, 3.25, 2.75, 1.5]],
        ["line", "black", [5, 3, 8, 3]],
        ["line", "black", [5,-3, 8,-3]],
    ],
}

class Vehicle(Robot):
    def __init__(
        self,
        x=0,
        y=0,
        a=0,
        color="white",
        name="Vickie",
        do_trace=True,
        height=0.25,
        max_trace_length=10,
        **kwargs
    ):
        config = {
            "x": x,
            "y": y,
            "a": a, # degrees in config file
            "color": color,
            "name": name,
            "do_trace": do_trace,
            "height": height,
            "max_trace_length": max_trace_length,
        }
        # First, get Scribbler defaults:
        defaults = VEHICLE_CONFIG.copy()
        # Then update from args:
        defaults.update(config)
        # Then update from kwargs:
        defaults.update(kwargs)
        super().__init__(**defaults)
